# coding=utf-8
#
# Copyright (C) 2016 Intel Corporation
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice (including the next
# paragraph) shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

"""Generate a set of tests to verify that formats added by the
 NV_image_formats extension for GLES 3.1 are properly accepted by the GLSL
 compiler.

 This also verifies that 16bits normalized formats are not accepted unless
 EXT_texture_norm16 is available.

 This program outputs, to stdout, the name of each file it generates.
"""

import os.path
from textwrap import dedent

from mako.template import Template

from modules import utils

def gen_header(status, norm16):
    """
    Generate a GLSL program header.

    Generate header code for ARB_shader_image_load_store GLSL parser
    tests that are expected to give status as result.
    """

    if norm16:
        norm16_require = "GL_EXT_texture_norm16"
    else:
        norm16_require = "!GL_EXT_texture_norm16"

    return dedent("""
        /*
         * [config]
         * expect_result: {0}
         * glsl_version: 3.10 es
         * require_extensions: GL_NV_image_formats {1}
         * [end config]
         */
        #version 310 es
        #extension GL_NV_image_formats : require
    """.format(status, norm16_require))


def gen(name, src, tests):
    """
    Expand a source template for the provided list of test definitions.

    Generate a GLSL parser test for each of the elements of the
    'tests' iterable, each of them should be a dictionary of
    definitions that will be used as environment to render the source
    template.

    The file name of each test will be the concatenation of the 'name'
    argument with the 'name' item from the respective test dictionary.
    """
    template = Template(dedent(src))

    for t in product([{'name': name}], tests):
        filename = os.path.join('spec',
                                'nv_image_formats',
                                'compiler',
                                '{0}.{1}'.format(t['name'],
                                                 t['shader_stage']))
        print(filename)

        dirname = os.path.dirname(filename)
        utils.safe_makedirs(dirname)

        with open(filename, 'w') as f:
            f.write(template.render(header=gen_header, **t))


shader_stages = [
    {'shader_stage': 'frag'},
    {'shader_stage': 'vert'}
]


image_types = [
    {
        'name': '2d',
        'image_type': 'image2D',
    },
    {
        'name': '3d',
        'image_type': 'image3D',
    },
    {
        'name': '2d-array',
        'image_type': 'image2DArray',
    },
    {
        'name': 'cube',
        'image_type': 'imageCube',
    }
]


def product(ps, *qss):
    """
    Generate the cartesian product of a number of lists of dictionaries.

    Each generated element will be the union of some combination of
    elements from the iterable arguments.  The resulting value of each
    'name' item will be the concatenation of names of the respective
    element combination separated with dashes.
    """
    for q in (product(*qss) if qss else [{}]):
        for p in ps:
            r = dict(p, **q)
            r['name'] = '-'.join(s['name'] for s in (p, q) if s.get('name'))
            yield r


def main():
    """Main function."""


    #
    # Test image formats accepted with NV_image_formats &
    # EXT_texture_norm16.
    #
    gen('declarations-with-norm16', """
        ${header('pass', True)}

        layout(rg32f) readonly uniform highp ${image_type} img_rg32f;
        layout(rg16f) readonly uniform highp ${image_type} img_rg16f;
        layout(r16f) readonly uniform highp ${image_type} img_r16f;
        layout(r11f_g11f_b10f) readonly uniform highp ${image_type} img_r11g11b10f;

        layout(rgba16) readonly uniform highp ${image_type} img_rgba16_unorm;
        layout(rgb10_a2) readonly uniform highp ${image_type} img_rgb10_a2_unorm;
        layout(rg16) readonly uniform highp ${image_type} img_rg16_unorm;
        layout(rg8) readonly uniform highp ${image_type} img_rg8_unorm;
        layout(r16) readonly uniform highp ${image_type} img_r16_unorm;
        layout(r8) readonly uniform highp ${image_type} img_r8_unorm;

        layout(rgba16_snorm) readonly uniform highp ${image_type} img_rgba16_snorm;
        layout(rg16_snorm) readonly uniform highp ${image_type} img_rg16_snorm;
        layout(rg8_snorm) readonly uniform highp ${image_type} img_rg8_snorm;
        layout(r16_snorm) readonly uniform highp ${image_type} img_r16_snorm;
        layout(r8_snorm) readonly uniform highp ${image_type} img_r8_snorm;

        layout(rgb10_a2ui) readonly uniform highp u${image_type} img_rgb10_a2ui;
        layout(rg32ui) readonly uniform highp u${image_type} img_rg32ui;
        layout(rg16ui) readonly uniform highp u${image_type} img_rg16ui;
        layout(rg8ui) readonly uniform highp u${image_type} img_rg8ui;
        layout(r16ui) readonly uniform highp u${image_type} img_r16ui;
        layout(r8ui) readonly uniform highp u${image_type} img_r8ui;

        layout(rg32i) readonly uniform highp i${image_type} img_rg32i;
        layout(rg16i) readonly uniform highp i${image_type} img_rg16i;
        layout(rg8i) readonly uniform highp i${image_type} img_rg8i;
        layout(r16i) readonly uniform highp i${image_type} img_r16i;
        layout(r8i) readonly uniform highp i${image_type} img_r8i;

        void main()
        {
        }
    """, product(image_types, shader_stages))


    #
    # Test image formats accepted with only NV_image_formats.
    #
    gen('declarations-without-norm16', """
        ${header('pass', False)}

        layout(rg32f) readonly uniform highp ${image_type} img_rg32f;
        layout(rg16f) readonly uniform highp ${image_type} img_rg16f;
        layout(r16f) readonly uniform highp ${image_type} img_r16f;
        layout(r11f_g11f_b10f) readonly uniform highp ${image_type} img_r11g11b10f;

        layout(rgb10_a2) readonly uniform highp ${image_type} img_rgb10_a2_unorm;
        layout(rg8) readonly uniform highp ${image_type} img_rg8_unorm;
        layout(r8) readonly uniform highp ${image_type} img_r8_unorm;

        layout(rg8_snorm) readonly uniform highp ${image_type} img_rg8_snorm;
        layout(r8_snorm) readonly uniform highp ${image_type} img_r8_snorm;

        layout(rgb10_a2ui) readonly uniform highp u${image_type} img_rgb10_a2ui;
        layout(rg32ui) readonly uniform highp u${image_type} img_rg32ui;
        layout(rg16ui) readonly uniform highp u${image_type} img_rg16ui;
        layout(rg8ui) readonly uniform highp u${image_type} img_rg8ui;
        layout(r16ui) readonly uniform highp u${image_type} img_r16ui;
        layout(r8ui) readonly uniform highp u${image_type} img_r8ui;

        layout(rg32i) readonly uniform highp i${image_type} img_rg32i;
        layout(rg16i) readonly uniform highp i${image_type} img_rg16i;
        layout(rg8i) readonly uniform highp i${image_type} img_rg8i;
        layout(r16i) readonly uniform highp i${image_type} img_r16i;
        layout(r8i) readonly uniform highp i${image_type} img_r8i;

        void main()
        {
        }
    """, product(image_types, shader_stages))


    #
    # Test image formats forbidden without EXT_texture_norm16.
    #
    gen('declaration-disallow-rgba16-unorm', """
        ${header('fail', False)}

        layout(rgba16) readonly uniform highp ${image_type} img_rgba16;

        void main()
        {
        }
    """, product(image_types, shader_stages))


    gen('declaration-disallow-rg16-unorm', """
        ${header('fail', False)}

        layout(rg16) readonly uniform highp ${image_type} img_rg16;

        void main()
        {
        }
    """, product(image_types, shader_stages))


    gen('declaration-disallow-r16-unorm', """
        ${header('fail', False)}

        layout(r16) readonly uniform highp ${image_type} img_r16;

        void main()
        {
        }
    """, product(image_types, shader_stages))


    gen('declaration-disallow-rgba16-snorm', """
        ${header('fail', False)}

        layout(rgba16_snorm) readonly uniform highp ${image_type} img_rgba16_snorm;

        void main()
        {
        }
    """, product(image_types, shader_stages))


    gen('declaration-disallow-rg16-snorm', """
        ${header('fail', False)}

        layout(rg16_snorm) readonly uniform highp ${image_type} img_rg16_snorm;

        void main()
        {
        }
    """, product(image_types, shader_stages))


    gen('declaration-disallow-r16-snorm', """
        ${header('fail', False)}

        layout(r16_snorm) readonly uniform highp ${image_type} img_r16_snorm;

        void main()
        {
        }
    """, product(image_types, shader_stages))

if __name__ == '__main__':
    main()



# coding=utf-8
# Copyright (c) 2016 Broadcom
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice (including the next
# paragraph) shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

from __future__ import (
    absolute_import, division, print_function, unicode_literals
)
import errno
import re
import subprocess

__all__ = [
    'DriverClassifier',
]

class DriverClassifier(object):
    def __init__(self):
        self.categories = []

        self.collect_driver_info()
        self.find_categories()


    def collect_driver_info(self):
        """Method for collecting the GL driver renderer string.

        Currently only glxinfo is used.
        """
        self.renderer = ''

        self.collect_glxinfo()


    def collect_glxinfo(self):
        """Calls glxinfo and parses the output to find the GL driver

        vendor/renderer strings.
        """

        try:
            output = subprocess.check_output(
                ['glxinfo'], stderr=subprocess.STDOUT).decode('utf-8')
        except OSError as e:
            if e.errno not in [errno.ENOENT, errno.EACCES]:
                raise
            return
        except subprocess.CalledProcessError:
            return

        for line in output.splitlines():
            m = re.match('OpenGL renderer string: (.*)', line)
            if m is not None:
                self.renderer = m.group(1)
                break


    def find_categories(self):
        """Parses the vendor/renderer strings to decide what categories

        the driver falls under.
        """
        if self.renderer.startswith(('Mesa ', 'Gallium ')):
            self.categories.append('mesa')

            m = re.match('.* VC4(.*)', self.renderer)
            if m is not None:
                self.categories.append('vc4')
                m = re.match(' V3D ([0-9])+\.([0-9])+', m.group(1))
                if m is not None:
                    self.categories.append('vc4-{}.{}'.format(m.group(1),
                                                              m.group(2)))

            m = re.match('Mesa DRI R200 ', self.renderer)
            if m is not None:
                self.categories.append('r200')

            m = re.match('Mesa DRI Intel[^ ]* (.*)', self.renderer)
            if m is not None:
                tail = m.group(1)

                i965_chipdict = {
                    '965': 'brw',
                    '946': 'brw',
                    '.*[GQ]4[35]': 'g4x',
                    'Ironlake': 'ilk',
                    'Sandybridge': 'snb',
                    'Ivybridge': 'ivb',
                    'Haswell': 'hsw',
                    'Baytrail': 'byt',
                    'Broadwell': 'bdw',
                    'Skylake': 'skl',
                    'HD Graphics .* \(Skylake': 'skl',
                    'Kabylake': 'kbl',
                    '.*Cherryview': 'chv',
                    '.*Broxton': 'bxt',
                }

                for chip, abbrev in i965_chipdict.items():
                    m = re.match(chip, tail)
                    if m is not None:
                        self.categories.append('i965')
                        self.categories.append('i965-{}'.format(abbrev))
                        break

        self.categories.reverse()



# coding=utf-8
# Copyright (c) 2014, 2016 Intel Corporation

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Provides tests for external integration with piglit.

These tests are by no means comprehensive, nor are they meant to be. The goal
is mainly just a sanity check to make sure the modules don't contain syntax
errors and to ensure that the API hasn't changed without fixing these modules
"""

from __future__ import (
    absolute_import, division, print_function, unicode_literals
)
import importlib

import pytest

from framework import core
from framework import exceptions


def setup_module():
    core.PIGLIT_CONFIG = core.PiglitConfig(allow_no_value=True)
    core.get_config()


def _import(name):
    """Helper for importing modules.

    It is very important that we use import_module to get the module, since we
    need more than just the profile, since we want to ensure that the Test
    derived class is importable.
    """
    try:
        return importlib.import_module(name)
    except exceptions.PiglitFatalError:
        pytest.skip('The module experienced a fatal error. '
                    'This may be expected.')


@pytest.mark.parametrize("name", [
    'tests.cts_gl',
    'tests.cts_gles',
    'tests.deqp_gles2',
    'tests.deqp_gles3',
    'tests.deqp_gles31',
    'tests.deqp_vk',
    'tests.deqp_egl',
    'tests.igt',
    'tests.oglconform',
    'tests.xts',
    'tests.xts-render',
])
def test_import(name):
    """Try to import tests, if they raise an error skip."""
    # TODO: Figure out what is necissary for each test to run, or catch
    # specific errors as skips rather than any of them.
    _import(name)


def test_xts_xtstest():
    """ xts.XTSTest initializes """
    mod = _import('tests.xts')
    mod.XTSTest('name', 'testname', 'testnum')


def test_xts_xtsprofile():
    """ xts.XTSProfile initializes """
    mod = _import('tests.xts')
    mod.XTSProfile()


def test_igt_igttest():
    """ igt.IGTTest initializes """
    mod = _import('tests.igt')
    mod.IGTTest('foo')


def test_oglconform_oglctest():
    """ oglconform.OGLCTest initializes """
    mod = _import('tests.oglconform')
    mod.OGLCTest('catagory', 'subtest')



# encoding=utf-8
# Copyright © 2018 Intel Corporation

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Script for taking profiles in python format and serializing them to XML."""

import argparse
import gzip
import os
import sys
import xml.etree.cElementTree as et

import six

sys.path.insert(0, os.path.join(os.path.dirname(os.path.abspath(__file__)), '..'))

from framework.test.piglit_test import (
    PiglitGLTest, PiglitCLTest, ASMParserTest, BuiltInConstantsTest,
    CLProgramTester, VkRunnerTest
)
from framework.test.shader_test import ShaderTest, MultiShaderTest
from framework.test.glsl_parser_test import GLSLParserTest
from framework.profile import load_test_profile
from framework.options import OPTIONS


def parser():
    """Parse command line arguments."""
    parser = argparse.ArgumentParser()
    parser.add_argument('name')
    parser.add_argument('input')
    parser.add_argument('output')
    parser.add_argument('--no-process-isolation', action='store_true')
    parser.add_argument('--glsl-arb-compat', action='store_true')
    args = parser.parse_args()
    return args


def _serialize_skips(test, elem):
    elems = [
        ('require_shader', 'shader_version'),
        ('require_api', 'api'),
        ('require_version', 'api_version'),
        ('require_extensions', 'extensions'),
    ]
    for e, f in elems:
        value = getattr(test, e, None)

        # For most tests (except MultiShader, we use the test value 'e', but
        # for MultiShader what's exposed is a FastSkip object, and the methods
        # are 'f'
        if not value:
            value = getattr(test, f, None)
        if value:
            et.SubElement(elem, 'option', name=f, value=repr(value))


def serializer(name, profile, outfile):
    """Take each test in the profile and write it out into the xml."""
    # TODO: This is going to take a lot of memory
    root = et.Element('PiglitTestList', count=six.text_type(len(profile)),
                      name=name)
    for name, test in profile.itertests():
        if isinstance(test, PiglitGLTest):
            elem = et.SubElement(root, 'Test', type='gl', name=name)
            if test.require_platforms:
                et.SubElement(elem, 'option', name='require_platforms',
                              value=repr(test.require_platforms))
            if test.exclude_platforms:
                et.SubElement(elem, 'option', name='exclude_platforms',
                              value=repr(test.exclude_platforms))
            _serialize_skips(test, elem)
        elif isinstance(test, BuiltInConstantsTest):
            elem = et.SubElement(root, 'Test', type='gl_builtin', name=name)
        elif isinstance(test, GLSLParserTest):
            elem = et.SubElement(root, 'Test', type='glsl_parser', name=name)
            _serialize_skips(test, elem)
        elif isinstance(test, ASMParserTest):
            elem = et.SubElement(root, 'Test', type='asm_parser', name=name)
            et.SubElement(elem, 'option', name='type_',
                          value=repr(test.command[1]))
            et.SubElement(elem, 'option', name='filename',
                          value=repr(test.filename))
            continue
        elif isinstance(test, ShaderTest):
            elem = et.SubElement(root, 'Test', type='shader', name=name)
            _serialize_skips(test, elem)
        elif isinstance(test, MultiShaderTest):
            elem = et.SubElement(root, 'Test', type='multi_shader', name=name)
            et.SubElement(elem, 'option', name='prog', value=repr(test.prog))
            et.SubElement(elem, 'option', name='files', value=repr(test.files))
            et.SubElement(elem, 'option', name='subtests', value=repr(test.subtests))
            skips = et.SubElement(elem, 'Skips')
            for s in test.skips:
                skip = et.SubElement(skips, 'Skip')
                _serialize_skips(s, skip)
            continue
        elif isinstance(test, CLProgramTester):
            elem = et.SubElement(root, 'Test', type='cl_prog', name=name)
            et.SubElement(elem, 'option', name='filename',
                          value=repr(test.filename))
            continue
        elif isinstance(test, PiglitCLTest):
            elem = et.SubElement(root, 'Test', type='cl', name=name)
            et.SubElement(elem, 'option', name='command', value=repr(test._command))
            continue
        elif isinstance(test, VkRunnerTest):
            elem = et.SubElement(root, 'Test', type='vkrunner', name=name)
            et.SubElement(elem, 'option', name='filename',
                          value=repr(test.filename))
            continue
        else:
            continue

        et.SubElement(elem, 'option', name='command', value=repr(test._command))
        et.SubElement(elem, 'option', name='run_concurrent',
                      value=repr(test.run_concurrent))
        if test.cwd:
            et.SubElement(elem, 'option', name='cwd', value=test.cwd)
        if test.env:
            env = et.SubElement(elem, 'environment')
            for k, v in six.iteritems(test.env):
                et.SubElement(env, 'env', name=k, value=v)

    tree = et.ElementTree(root)
    with gzip.open(outfile, 'wb') as f:
        tree.write(f, encoding='utf-8', xml_declaration=True)


def main():
    args = parser()
    OPTIONS.process_isolation = not args.no_process_isolation
    if args.glsl_arb_compat:
        os.environ['PIGLIT_FORCE_GLSLPARSER_DESKTOP'] = 'true'
    profile = load_test_profile(args.input, python=True)
    serializer(args.name, profile, args.output)


if __name__ == '__main__':
    main()



# coding=utf-8
#
# Copyright © 2012, 2014 Intel Corporation
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice (including the next
# paragraph) shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

from __future__ import print_function, division, absolute_import
import os
import itertools

from templates import template_dir
from modules import utils

TEMPLATES = template_dir(os.path.basename(os.path.splitext(__file__)[0]))

_NAMES = {
    "++t": "preincrement",
    "--t": "predecrement",
    "t++": "postincrement",
    "t--": "postdecrement",
}


def generate(dirname, type_name, op, usage, shader_target):
    """Generate glsl parser tests."""
    if type_name.endswith('2'):
        var_as_vec4 = 'vec4(t.xyxy)'
        components = '.xy'
    elif type_name.endswith('3'):
        var_as_vec4 = 'vec4(t.xyzx)'
        components = '.xyz'
    elif type_name.endswith('4'):
        var_as_vec4 = 'vec4(t)'
        components = ''
    else:
        var_as_vec4 = 'vec4(t)'
        components = '.x'

    if shader_target == 'vert':
        dest = "gl_Position"
        mode = 'attribute'
    else:
        mode = 'varying'
        dest = "gl_FragColor"

    filename = os.path.join(
        dirname,
        '{0}-{1}-non-lvalue-for-{2}.{3}'.format(
            _NAMES[op], type_name, usage, shader_target))

    print(filename)
    with open(filename, 'w') as f:
        f.write(TEMPLATES.get_template(
            '{0}.glsl_parser_test.mako'.format(usage)).render_unicode(
                type_name=type_name,
                mode=mode,
                dest=dest,
                components=components,
                var_as_vec4=var_as_vec4,
                op=op))


def all_tests():
    type_name = ['float', 'vec2', 'vec3', 'vec4', 'int', 'ivec2', 'ivec3',
                 'ivec4']
    op = ["++t", "--t", "t++", "t--"]
    usage = ['assignment', 'out-parameter']
    shader_target = ['vert', 'frag']

    for t, o, u, s in itertools.product(type_name, op, usage, shader_target):
        yield t, o, u, s


def main():
    dirname = os.path.join('spec', 'glsl-1.10', 'compiler', 'expressions')
    utils.safe_makedirs(dirname)

    for args in all_tests():
        generate(dirname, *args)


if __name__ == '__main__':
    main()



# coding=utf-8
#
# Copyright © 2013, 2014 Intel Corporation
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice (including the next
# paragraph) shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

from __future__ import print_function, division, absolute_import
import os
import os.path

import six

from templates import template_file
from modules import utils

TEMPLATE = template_file(os.path.basename(os.path.splitext(__file__)[0]),
                         'template.glsl_parser_test.mako')

SAMPLER_TYPE_TO_COORD_TYPE = {
    'sampler1D':              'float',
    'isampler1D':             'float',
    'usampler1D':             'float',

    'sampler2D':              'vec2',
    'isampler2D':             'vec2',
    'usampler2D':             'vec2',

    'sampler3D':              'vec3',
    'isampler3D':             'vec3',
    'usampler3D':             'vec3',

    'samplerCube':            'vec3',
    'isamplerCube':           'vec3',
    'usamplerCube':           'vec3',

    'sampler1DArray':         'float',
    'isampler1DArray':        'float',
    'usampler1DArray':        'float',

    'sampler2DArray':         'vec2',
    'isampler2DArray':        'vec2',
    'usampler2DArray':        'vec2',

    'samplerCubeArray':       'vec3',
    'isamplerCubeArray':      'vec3',
    'usamplerCubeArray':      'vec3',

    'sampler1DShadow':        'float',
    'sampler2DShadow':        'vec2',
    'samplerCubeShadow':      'vec3',
    'sampler1DArrayShadow':   'float',
    'sampler2DArrayShadow':   'vec2',
    'samplerCubeArrayShadow': 'vec3',
}

REQUIREMENTS = {
    'ARB_texture_query_lod': {
        'version': '1.30',
        'extensions': 'GL_ARB_texture_query_lod'
    },
    'glsl-4.00': {
        'version': '4.00',
        'extensions': None
    }
}


def main():
    """Main function."""
    for api, requirement in six.iteritems(REQUIREMENTS):
        lod = 'Lod' if api == 'glsl-4.00' else 'LOD'
        dirname = os.path.join("spec", api.lower(), "compiler",
                               "built-in-functions")
        utils.safe_makedirs(dirname)

        for sampler_type, coord_type in six.iteritems(SAMPLER_TYPE_TO_COORD_TYPE):
            requirements = [requirement['extensions']] if requirement['extensions'] else []

            # samplerCubeArray types are part GLSL 4.00
            # or GL_ARB_texture_cube_map_array.
            if api == "ARB_texture_query_lod" and sampler_type in [
                    'samplerCubeArray', 'isamplerCubeArray',
                    'usamplerCubeArray', 'samplerCubeArrayShadow']:
                requirements.append('GL_ARB_texture_cube_map_array')

            for execution_stage in ("vs", "fs"):
                file_extension = 'frag' if execution_stage == 'fs' else 'vert'
                filename = os.path.join(
                    dirname,
                    "textureQuery{0}-{1}.{2}".format(lod, sampler_type,
                                                     file_extension))
                print(filename)

                with open(filename, "w") as f:
                    f.write(TEMPLATE.render_unicode(
                        version=requirement['version'],
                        extensions=requirements,
                        execution_stage=execution_stage,
                        sampler_type=sampler_type,
                        coord_type=coord_type,
                        lod=lod))


if __name__ == '__main__':
    main()



# coding=utf-8
# Copyright (c) 2015-2016 Intel Corporation

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Test the opengl module."""

from __future__ import (
    absolute_import, division, print_function, unicode_literals
)
try:
    from unittest import mock
except ImportError:
    import mock

import pytest

from framework import wflinfo
from framework.test import opengl
from framework.test.base import TestIsSkip as _TestIsSkip

from .. import utils

# pylint: disable=no-self-use,attribute-defined-outside-init,protected-access


class TestFastSkipMixin(object):  # pylint: disable=too-many-public-methods
    """Tests for the FastSkipMixin class."""

    @pytest.yield_fixture(autouse=True, scope='class')
    def patch(self):
        """Create a Class with FastSkipMixin, but patch various bits."""
        _mock_wflinfo = mock.Mock(spec=wflinfo.WflInfo)
        _mock_wflinfo.core.api_version = 3.3
        _mock_wflinfo.core.shader_version = 3.3
        _mock_wflinfo.core.extensions = set(['bar'])
        _mock_wflinfo.es2.api_version = 3.0
        _mock_wflinfo.es2.shader_version = 2.0
        _mock_wflinfo.es2.extensions = set(['bar'])

        with mock.patch('framework.test.opengl.FastSkip.info', _mock_wflinfo):
            yield

    class _Test(opengl.FastSkipMixin, utils.Test):
        pass

    def test_api(self):
        """Tests that the api works.

        Since you're not suppoed to be able to pass gl and gles version, this
        uses two seperate constructor calls.
        """
        self._Test(['foo'], extensions={'foo'}, api_version=3, shader_version=2)
        self._Test(['foo'], extensions={'foo'}, api_version=3,
                   shader_version=2)


class TestFastSkip(object):
    """Tests for the FastSkip class."""

    @pytest.yield_fixture(autouse=True, scope='class')
    def patch(self):
        """Create a Class with FastSkipMixin, but patch various bits."""
        _mock_wflinfo = mock.Mock(spec=wflinfo.WflInfo)
        _mock_wflinfo.core.api_version = 3.3
        _mock_wflinfo.core.shader_version = 3.3
        _mock_wflinfo.core.extensions = set(['bar'])
        _mock_wflinfo.es2.api_version = 3.0
        _mock_wflinfo.es2.shader_version = 2.0
        _mock_wflinfo.es2.extensions = set(['bar'])

        with mock.patch('framework.test.opengl.FastSkip.info', _mock_wflinfo):
            yield

    @pytest.fixture
    def inst(self):
        return opengl.FastSkip(api='core')

    def test_should_skip(self, inst):
        """test.opengl.FastSkipMixin.test: Skips when requires is missing
        from extensions.
        """
        inst.extensions.add('foobar')
        with pytest.raises(_TestIsSkip):
            inst.test()

    def test_should_not_skip(self, inst):
        """test.opengl.FastSkipMixin.test: runs when requires is in
        extensions.
        """
        inst.extensions.add('bar')
        inst.test()

    def test_max_api_version_lt(self, inst):
        """test.opengl.FastSkipMixin.test: skips if api_version >
        __max_api_version.
        """
        inst.api_version = 4.0
        with pytest.raises(_TestIsSkip):
            inst.test()

    def test_max_api_version_gt(self, inst):
        """test.opengl.FastSkipMixin.test: runs if api_version <
        __max_api_version.
        """
        inst.api_version = 1.0

    def test_max_api_version_set(self, inst):
        """test.opengl.FastSkipMixin.test: runs if api_version is None"""
        inst.test()

    def test_max_api_version_lt(self, inst):
        """test.opengl.FastSkipMixin.test: skips if api_version >
        __max_api_version.
        """
        inst.api_version = 4.0
        with pytest.raises(_TestIsSkip):
            inst.test()

    def test_max_api_version_gt(self, inst):
        """test.opengl.FastSkipMixin.test: runs if api_version <
        __max_api_version.
        """
        inst.api_version = 1.0

    def test_max_api_version_set(self, inst):
        """test.opengl.FastSkipMixin.test: runs if api_version is None"""
        inst.test()

    def test_max_shader_version_lt(self, inst):
        """test.opengl.FastSkipMixin.test: skips if shader_version >
        __max_shader_version.
        """
        inst.shader_version = 4.0
        with pytest.raises(_TestIsSkip):
            inst.test()

    def test_max_shader_version_gt(self, inst):
        """test.opengl.FastSkipMixin.test: runs if shader_version <
        __max_shader_version.
        """
        inst.shader_version = 1.0

    def test_max_shader_version_set(self, inst):
        """test.opengl.FastSkipMixin.test: runs if shader_version is None"""
        inst.test()

    def test_max_shader_version_lt(self, inst):
        """test.opengl.FastSkipMixin.test: skips if shader_version >
        __max_shader_version.
        """
        inst.shader_version = 4.0
        with pytest.raises(_TestIsSkip):
            inst.test()

    def test_max_shader_version_gt(self, inst):
        """test.opengl.FastSkipMixin.test: runs if shader_version <
        __max_shader_version.
        """
        inst.shader_version = 1.0

    def test_max_shader_version_set(self, inst):
        """test.opengl.FastSkipMixin.test: runs if shader_version is None"""
        inst.test()

    class TestEmpty(object):
        """Tests for the FastSkip class when values are unset."""

        @pytest.yield_fixture(autouse=True, scope='class')
        def patch(self):
            """Create a Class with FastSkipMixin, but patch various bits."""
            _mock_wflinfo = mock.Mock(spec=wflinfo.WflInfo)
            _mock_wflinfo.core.api_version = 0.0
            _mock_wflinfo.core.shader_version = 0.0
            _mock_wflinfo.core.extensions = set()
            _mock_wflinfo.es2.api_version = 0.0
            _mock_wflinfo.es2.shader_version = 0.0
            _mock_wflinfo.es2.extensions = set()

            with mock.patch('framework.test.opengl.FastSkip.info', _mock_wflinfo):
                yield

        @pytest.fixture
        def inst(self):
            return opengl.FastSkip(api='core')

        def test_extension_empty(self, inst):
            """test.opengl.FastSkipMixin.test: if extensions are empty test
            runs.
            """
            inst.info.core.extensions.add('foobar')
            inst.test()

        def test_requires_empty(self, inst):
            """test.opengl.FastSkipMixin.test: if gl_requires is empty test
            runs.
            """
            inst.test()

    def test_max_shader_version_unset(self, inst):
        """test.opengl.FastSkipMixin.test: runs if __max_shader_version is
        None.
        """
        inst.shader_version = 1.0
        inst.test()

    def test_max_api_version_unset(self, inst):
        """test.opengl.FastSkipMixin.test: runs if __max_api_version is
        None.
        """
        inst.api_version = 1.0
        inst.test()


class TestFastSkipMixinDisabled(object):
    """Tests for the sub version."""

    @pytest.yield_fixture(autouse=True, scope='class')
    def patch(self):
        """Create a Class with FastSkipMixin, but patch various bits."""
        _mock_wflinfo = mock.Mock(spec=wflinfo.WflInfo)
        _mock_wflinfo.es2.api_version = 3.3
        _mock_wflinfo.es2.shader_version = 2.0
        _mock_wflinfo.es2.extensions = set(['bar'])
        _mock_wflinfo.core.api_version = 3.0
        _mock_wflinfo.core.shader_version = 3.3
        _mock_wflinfo.core.extensions = set(['bar'])

        with mock.patch('framework.test.opengl.FastSkip.info', _mock_wflinfo):
            yield

    class _Test(opengl.FastSkipMixin, utils.Test):
        pass

    def test_api(self):
        """Tests that the api works.

        Since you're not suppoed to be able to pass gl and gles version, this
        uses two seperate constructor calls.
        """
        self._Test(['foo'], extensions={'foo'}, api_version=3, shader_version=2)
        self._Test(['foo'], extensions={'foo'}, api_version=3,
                   shader_version=2)



# coding=utf-8
#
# Copyright (C) 2016 Intel Corporation
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice (including the next
# paragraph) shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

"""Generate a set of tests to verify that layout qualifiers introduced
 by INTEL_conservative_rasterization are properly parsed.

 This also verifies the interaction with ARB_post_depth_coverage.

 This program outputs, to stdout, the name of each file it generates.

"""

import os.path
from textwrap import dedent

from mako.template import Template

from modules import utils

def gen_header(status, gl_api, shader_stage):
    """
    Generate a GLSL program header.

    Generate header code for INTEL_conservative_rasterization GLSL parser
    tests that are expected to give status as result.
    """

    if shader_stage != 'frag':
        status = 'fail'
    print("%s - %s" % (shader_stage, status))

    if gl_api == "gles":
        glsl_version = ("3.20 es", "320 es")
    else:
        glsl_version = ("4.20", "420")

    return dedent("""
        /*
         * [config]
         * expect_result: {0}
         * glsl_version: {1}
         * require_extensions: GL_INTEL_conservative_rasterization
         * [end config]
         */
        #version {2}
        #extension GL_INTEL_conservative_rasterization : enable
    """.format(status, glsl_version[0], glsl_version[1]))


def gen(name, src, tests):
    """
    Expand a source template for the provided list of test definitions.

    Generate a GLSL parser test for each of the elements of the
    'tests' iterable, each of them should be a dictionary of
    definitions that will be used as environment to render the source
    template.

    The file name of each test will be the concatenation of the 'name'
    argument with the 'name' item from the respective test dictionary.
    """
    template = Template(dedent(src))

    for t in product([{'name': name}], tests):
        filename = os.path.join('spec',
                                'intel_conservative_rasterization',
                                'compiler',
                                '{0}.{1}.{2}'.format(t['name'],
                                                     t['gl_api'],
                                                     t['shader_stage']))
        print(filename)

        dirname = os.path.dirname(filename)
        utils.safe_makedirs(dirname)

        with open(filename, 'w') as f:
            f.write(template.render(header=gen_header, **t))


shader_stages = [
    {'shader_stage': 'frag'},
    {'shader_stage': 'vert'}
]


gl_apis = [
    {'gl_api': 'gl'},
    {'gl_api': 'gles'}
]


def product(ps, *qss):
    """
    Generate the cartesian product of a number of lists of dictionaries.

    Each generated element will be the union of some combination of
    elements from the iterable arguments.  The resulting value of each
    'name' item will be the concatenation of names of the respective
    element combination separated with dashes.
    """
    for q in (product(*qss) if qss else [{}]):
        for p in ps:
            r = dict(p, **q)
            r['name'] = '-'.join(s['name'] for s in (p, q) if s.get('name'))
            yield r


def main():
    """Main function."""


    #
    # Test inner_coverage layout qualifier.
    #
    gen('inner_coverage', """
        ${header('pass', gl_api, shader_stage)}

        layout(inner_coverage) in;

        void main()
        {
        }
    """, product(gl_apis, shader_stages))

    #
    # Test depth_coverage layout qualifier.
    #
    gen('post_depth_coverage', """
        ${header('pass', gl_api, shader_stage)}

        layout(post_depth_coverage) in;

        void main()
        {
        }
    """, product(gl_apis, shader_stages))

    #
    # Test depth_coverage layout qualifier.
    #
    gen('inner_post_depth_coverage', """
        ${header('fail', gl_api, shader_stage)}

        layout(inner_coverage) in;
        layout(post_depth_coverage) in;

        void main()
        {
        }
    """, product(gl_apis, shader_stages))

if __name__ == '__main__':
    main()



# coding=utf-8
# Copyright 2016 Advanced Micro Devices, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice (including the next
# paragraph) shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import print_function, division, absolute_import
import os
import random
import textwrap

from six.moves import range

from modules import utils

TYPES = ['char', 'uchar', 'short', 'ushort', 'int', 'uint', 'long', 'ulong', 'half', 'float', 'double']
VEC_SIZES = ['2', '3', '4', '8', '16']

DIR_NAME = os.path.join("cl", "vstore")


def gen_array(size):
    random.seed(size)
    return ' '.join([str(random.randint(0, 255)) for i in range(size)])


def ext_req(type_name):
    if type_name[:6] == "double":
        return "require_device_extensions: cl_khr_fp64"
    if type_name[:4] == "half":
        return "require_device_extensions: cl_khr_fp16"
    return ""


def begin_test(suffix, type_name, mem_type, vec_sizes, addr_space, aligned):
    file_name = os.path.join(DIR_NAME, "vstore{}-{}-{}.cl".format(suffix, type_name, addr_space))
    print(file_name)
    f = open(file_name, 'w')
    f.write(textwrap.dedent(("""\
    /*!
    [config]
    name: Vector store{suffix} {addr_space} {type_name}2,3,4,8,16
    clc_version_min: 11

    dimensions: 1
    global_size: 1 0 0
    """ + ext_req(type_name))
    .format(type_name=type_name, addr_space=addr_space, suffix=suffix)))
    for s in vec_sizes:
        size = int(s) if s != '' else 1
        modsize = 4 if size == 3 and aligned else size
        offset = modsize if aligned else 1
        canary= '0xdeadp1' if type_name in ('float', 'double') else '0xdead'

        ty_name = type_name + s
        f.write(textwrap.dedent("""
        [test]
        name: vector store{suffix} {addr_space} {type_name}
        kernel_name: vstore{suffix}{n}_{addr_space}
        arg_out: 0 buffer {mem_type}[{size}] {offset_zeros}{gen_array} {canary}
        arg_in: 0 buffer {mem_type}[{size}] {offset_size_zeros} {canary}
        arg_in:  1 buffer {type_name}[1] {gen_array}

        [test]
        name: vector store{suffix} {addr_space} offset {type_name}
        kernel_name: vstore{suffix}{n}_{addr_space}_offset
        arg_out: 0 buffer {mem_type}[{offset_size}] {offset_zeros} {gen_array} {padd_zeros} {gen_array} {canary}
        arg_in: 0 buffer {mem_type}[{offset_size}] {offset_modsize_size_zeros} {canary}
        arg_in: 1 buffer {type_name}[1] {gen_array}
        """.format(type_name=ty_name, mem_type=mem_type, size=size + offset + 1,
                   offset_zeros = ("0 " * offset),
                   offset_size_zeros = ("0 " * (offset + size)),
                   padd_zeros = ("0 " * (modsize - size)),
                   offset_modsize_size_zeros = ("0 " * (modsize + size + offset)),
                   offset_size=modsize + size + offset + 1, n=s,
                   gen_array=gen_array(size),
                   suffix=suffix, addr_space=addr_space,
                   canary=canary)))

    f.write(textwrap.dedent("""
    !*/
    """))
    if type_name == "double":
        f.write(textwrap.dedent("""
        #pragma OPENCL EXTENSION cl_khr_fp64: enable
        """))
    if type_name == "half":
        f.write(textwrap.dedent("""
        #pragma OPENCL EXTENSION cl_khr_fp16: enable
        """))
    return f


def gen_test_global(suffix, t, mem_type, vec_sizes, aligned):
    f = begin_test(suffix, t, mem_type, vec_sizes, 'global', aligned)
    for s in vec_sizes:
        offset = int(s) if aligned else 1
        offset = 4 if offset == 3 else offset

        type_name = t + s
        f.write(textwrap.dedent("""
        kernel void vstore{suffix}{n}_global(global {mem_type} *out,
                                     global {type_name} *in) {{
            {type_name} tmp = in[0];
            vstore{suffix}{n}(tmp, 0, out + {offset});
        }}

        kernel void vstore{suffix}{n}_global_offset(global {mem_type} *out,
                                            global {type_name} *in) {{
            {type_name} tmp = in[0];
            vstore{suffix}{n}(tmp, 0, out + {offset});
            vstore{suffix}{n}(tmp, 1, out + {offset});
        }}
        """.format(type_name=type_name, mem_type=mem_type, n=s, suffix=suffix,
                   offset=offset)))

    f.close()


def gen_test_local_private(suffix, t, mem_type, vec_sizes, addr_space, aligned):
    f = begin_test(suffix, t, mem_type, vec_sizes, addr_space, aligned)
    for s in vec_sizes:
        size = int(s) if s != '' else 1
        modsize = 4 if size == 3 and aligned else size
        offset = modsize if aligned else 1

        type_name = t + s
        f.write(textwrap.dedent("""
        kernel void vstore{suffix}{n}_{addr_space}(global {mem_type} *out,
                                     global {type_name} *in) {{
            {type_name} tmp = in[0];
            volatile {addr_space} {mem_type} loc[{size}];
            for (int i = 0; i < {size}; ++i)
                loc[i] = ({mem_type})0;

            vstore{suffix}{n}(tmp, 0, ({addr_space} {mem_type}*)loc + {offset});
            for (int i = 0; i < {size}; ++i)
                out[i] = loc[i];
        }}

        kernel void vstore{suffix}{n}_{addr_space}_offset(global {mem_type} *out,
                                            global {type_name} *in) {{
            {type_name} tmp = in[0];
            volatile {addr_space} {mem_type} loc[{offset_size}];
            for (int i = 0; i < {offset_size}; ++i)
                loc[i] = ({mem_type})0;

            vstore{suffix}{n}(tmp, 0, ({addr_space} {mem_type}*)loc + {offset});
            vstore{suffix}{n}(tmp, 1, ({addr_space} {mem_type}*)loc + {offset});
            for (int i = 0; i < {offset_size}; ++i)
                out[i] = loc[i];
        }}
        """.format(type_name=type_name, mem_type=mem_type, n=s, suffix=suffix,
                   offset_size=size + modsize + offset, size=size + offset,
                   addr_space=addr_space, offset=offset)))

    f.close()


# vstore_half is special, because CLC won't allow us to use half type without
# cl_khr_fp16
def gen_test_local_private_half(suffix, t, vec_sizes, addr_space, aligned):
    f = begin_test(suffix, t, 'half', vec_sizes, addr_space, aligned)
    for s in vec_sizes:
        size = int(s) if s != '' else 1
        modsize = 4 if size == 3 and aligned else size
        offset = modsize if aligned else 1

        type_name = t + s
        f.write(textwrap.dedent("""
        kernel void vstore{suffix}{n}_{addr_space}(global half *out,
                                     global {type_name} *in) {{
            {type_name} tmp = in[0];
            volatile {addr_space} short loc[{size}];
            for (int i = 0; i < {size}; ++i)
                loc[i] = 0;

            vstore{suffix}{n}(tmp, 0, ({addr_space} half*)loc + {offset});

            for (int i = 0; i < {size}; ++i)
                ((global short *)out)[i] = loc[i];
        }}

        kernel void vstore{suffix}{n}_{addr_space}_offset(global half *out,
                                            global {type_name} *in) {{
            {type_name} tmp = in[0];
            volatile {addr_space} short loc[{offset_size}];
            for (int i = 0; i < {offset_size}; ++i)
                loc[i] = 0;

            vstore{suffix}{n}(tmp, 0, ({addr_space} half*)loc + {offset});
            vstore{suffix}{n}(tmp, 1, ({addr_space} half*)loc + {offset});

            for (int i = 0; i < {offset_size}; ++i)
                ((global short *)out)[i] = loc[i];
        }}
        """.format(type_name=type_name, n=s, suffix=suffix,
                   offset_size=size + modsize + offset, size=size + offset,
                   addr_space=addr_space, offset=offset)))


def gen_test_local(suffix, t, mem_type, vec_sizes, aligned):
    if mem_type == 'half':
        gen_test_local_private_half(suffix, t, vec_sizes, 'local', aligned)
    else:
        gen_test_local_private(suffix, t, mem_type, vec_sizes, 'local', aligned)


def gen_test_private(suffix, t, mem_type, vec_sizes, aligned):
    if mem_type == 'half':
        gen_test_local_private_half(suffix, t, vec_sizes, 'private', aligned)
    else:
        gen_test_local_private(suffix, t, mem_type, vec_sizes, 'private', aligned)


def main():
    utils.safe_makedirs(DIR_NAME)
    for t in TYPES:
        gen_test_global('', t, t, VEC_SIZES, False);
        gen_test_local('', t, t, VEC_SIZES, False);
        gen_test_private('', t, t, VEC_SIZES, False);

    for aligned in False, True:
        suffix = "a_half" if aligned else "_half"
        vec_sizes = VEC_SIZES if aligned else [''] + VEC_SIZES

        gen_test_global(suffix, 'float',  'half', vec_sizes, aligned);
        gen_test_global(suffix, 'double', 'half', vec_sizes, aligned);
        gen_test_local(suffix, 'float',  'half', vec_sizes, aligned);
        gen_test_local(suffix, 'double', 'half', vec_sizes, aligned);
        gen_test_private(suffix, 'float',  'half', vec_sizes, aligned);
        gen_test_private(suffix, 'double', 'half', vec_sizes, aligned);


if __name__ == '__main__':
    main()



# coding=utf-8
#
# Copyright © 2011 Intel Corporation
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice (including the next
# paragraph) shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

# This source file defines a set of test vectors that can be used to
# test GLSL's built-in functions and operators.  It is intended to be
# used by Python code that generates Piglit tests.
#
# The key export is the dictionary test_suite.  It contains an entry
# for each possible overload of every pure built-in function and
# operator.  By iterating through this dictionary you can find a set
# of test vectors for testing nearly every built-in GLSL function.
#
# The following functions are not included, since they are not pure,
# so they can't be tested using simple vectors:
# - dFdx()
# - dFdy()
# - fwidth()
# - ftransform()
# - Increment and decrement operators
#
# The following functions are not included, since they need to be
# tested in specialized ways:
# - modf(): not tested because it has an out parameter
# - isnan() and isinf(): not tested because special effort is required
#   to create values that cause these functions to return true.
#
# Also not tested are array subscripting, field/method selection,
# swizzling, the function call operator, assignment, and the sequence
# operator.

from __future__ import print_function, division, absolute_import
import collections
import itertools
import functools

from six.moves import range
import numpy as np


# Floating point types used by Python and numpy
DOUBLE_TYPES = (float, np.float64, np.float32)

class GlslBuiltinType(object):
    """Class representing a GLSL built-in type."""
    def __init__(self, name, base_type, num_cols, num_rows,
                 version_introduced):
        self.__name = name
        if base_type is not None:
            self.__base_type = base_type
        else:
            self.__base_type = self
        self.__num_cols = num_cols
        self.__num_rows = num_rows
        self.__version_introduced = version_introduced

    @property
    def name(self):
        """The name of the type, as a string."""
        return self.__name

    @property
    def base_type(self):
        """For vectors and matrices, the type of data stored in each
        element.  For scalars, equal to self.
        """
        return self.__base_type

    @property
    def num_cols(self):
        """For matrices, the number of columns.  For vectors and
        scalars, 1.
        """
        return self.__num_cols

    @property
    def num_rows(self):
        """For vectors and matrices, the number of rows.  For scalars,
        1.
        """
        return self.__num_rows

    @property
    def is_scalar(self):
        return self.__num_cols == 1 and self.__num_rows == 1

    @property
    def is_vector(self):
        return self.__num_cols == 1 and self.__num_rows != 1

    @property
    def is_matrix(self):
        return self.__num_cols != 1

    @property
    def version_introduced(self):
        """The earliest version of GLSL that this type appears in (as
        a string, e.g. 110).
        """
        return self.__version_introduced

    def __eq__(self, other):
        if isinstance(other, GlslBuiltinType):
            return self.name == other.name

        return NotImplemented

    def __lt__(self, other):
        if isinstance(other, GlslBuiltinType):
            return self.name < other.name

        return NotImplemented

    def __hash__(self):
        """Hash the object.

        This hash isn't super awesome, but it isn't prone to change since you
        have to muck with private (__prefixed values) and some unlikely text in
        addition.

        """
        return hash('__GLslBuiltinType_{}__'.format(self.name))

    def __str__(self):
        return self.__name

    def __repr__(self):
        return 'glsl_{0}'.format(self.__name)


# Concrete declarations of GlslBuiltinType
glsl_bool   = GlslBuiltinType('bool',   None,       1, 1, 110)
glsl_bvec2  = GlslBuiltinType('bvec2',  glsl_bool,  1, 2, 110)
glsl_bvec3  = GlslBuiltinType('bvec3',  glsl_bool,  1, 3, 110)
glsl_bvec4  = GlslBuiltinType('bvec4',  glsl_bool,  1, 4, 110)
glsl_double = GlslBuiltinType('double', None,       1, 1, 400)
glsl_dvec2  = GlslBuiltinType('dvec2', glsl_double,  1, 2, 400)
glsl_dvec3  = GlslBuiltinType('dvec3', glsl_double,  1, 3, 400)
glsl_dvec4  = GlslBuiltinType('dvec4', glsl_double,  1, 4, 400)
glsl_dmat2   = GlslBuiltinType('dmat2', glsl_double, 2, 2, 400)
glsl_dmat3   = GlslBuiltinType('dmat3', glsl_double, 3, 3, 400)
glsl_dmat4   = GlslBuiltinType('dmat4', glsl_double, 4, 4, 400)
glsl_dmat2x2 = glsl_dmat2
glsl_dmat3x2 = GlslBuiltinType('dmat3x2', glsl_double, 3, 2, 400)
glsl_dmat4x2 = GlslBuiltinType('dmat4x2', glsl_double, 4, 2, 400)
glsl_dmat2x3 = GlslBuiltinType('dmat2x3', glsl_double, 2, 3, 400)
glsl_dmat3x3 = glsl_dmat3
glsl_dmat4x3 = GlslBuiltinType('dmat4x3', glsl_double, 4, 3, 400)
glsl_dmat2x4 = GlslBuiltinType('dmat2x4', glsl_double, 2, 4, 400)
glsl_dmat3x4 = GlslBuiltinType('dmat3x4', glsl_double, 3, 4, 400)
glsl_dmat4x4 = glsl_dmat4


# Named tuple representing the signature of a single overload of a
# built-in GLSL function or operator:
# - name is a name suitable for use in test filenames.  For functions,
#   this is the name of the function.  For operators, it is a short
#   description of the operator, beginning with "op", e.g. "op-plus".
# - template is a Python format string that can be used to construct
#   GLSL code that invokes the function or operator.
# - version_introduced earliest version of GLSL the test applies to
#   (as a string, e.g. 110).
# - rettype is the return type of the function or operator (as a
#   GlslBuiltinType).
# - argtypes is a tuple containing the types of each parameter (as
#   GlslBuiltinTypes).
#
# For example, the function
#
#   vec3 step(float edge, vec3 x)
#
# has a signature of
#
# Signature(name='step', template='step({0}, {1})',
#           version_introduced=110, rettype='vec3',
#           argtypes=('float', 'vec3'))
Signature = collections.namedtuple(
    'Signature',
    ('name', 'template', 'version_introduced', 'extension', 'rettype', 'argtypes'))


# Named tuple representing a single piece of test data for testing a
# built-in GLSL function:
# - arguments is a tuple containing the arguments to apply to the
#   function.  Each argument is of a type native to numpy (e.g.
#   numpy.float32 or numpy.ndarray)
# - result is the value the function is expected to return.  It is
#   also of a type native to numpy.
# - tolerance is a float64 representing how much deviation from the
#   result we expect, considering the floating point precision
#   requirements of GLSL and OpenGL.  The value may be zero for test
#   vectors involving booleans and integers.  If result is a vector or
#   matrix, tolerance should be interpreted as the maximum permissible
#   RMS error (as would be computed by the distance() function).
TestVector = collections.namedtuple(
    'TestVector', ('arguments', 'result', 'tolerance'))


def glsl_type_of(value):
    """Return the GLSL type corresponding to the given native numpy
    value, as a GlslBuiltinType.
    """

    if isinstance(value, DOUBLE_TYPES):
        return glsl_double
    elif isinstance(value, (bool, np.bool_)):
        return glsl_bool
    else:
        if len(value.shape) == 1:
            # Vector
            vector_length = value.shape[0]
            assert 2 <= vector_length <= 4
            if value.dtype in DOUBLE_TYPES:
                return (glsl_dvec2, glsl_dvec3, glsl_dvec4)[vector_length - 2]
            elif value.dtype == bool:
                return (glsl_bvec2, glsl_bvec3, glsl_bvec4)[vector_length - 2]
            else:
                raise Exception(
                    'Unexpected vector base type {0}'.format(value.dtype))
        else:
            # Matrix
            assert value.dtype in DOUBLE_TYPES
            assert len(value.shape) == 2
            matrix_rows = value.shape[0]
            assert 2 <= matrix_rows <= 4
            matrix_columns = value.shape[1]
            assert 2 <= matrix_columns <= 4
            matrix_types = ((glsl_dmat2x2, glsl_dmat2x3, glsl_dmat2x4),
                            (glsl_dmat3x2, glsl_dmat3x3, glsl_dmat3x4),
                            (glsl_dmat4x2, glsl_dmat4x3, glsl_dmat4x4))
            return matrix_types[matrix_columns - 2][matrix_rows - 2]


def column_major_values(value):
    """Given a native numpy value, return a list of the scalar values
    comprising it, in column-major order."""
    if isinstance(value, np.ndarray):
        return list(np.reshape(value, -1, 'F'))
    else:
        return [value]


def glsl_constant(value):
    """Given a native numpy value, return GLSL code that constructs
    it."""
    column_major = np.reshape(np.array(value), -1, 'F')
    if column_major.dtype == bool:
        values = ['true' if x else 'false' for x in column_major]
    else:
        values = ['{0}lf'.format(repr(x)) for x in column_major]
    if len(column_major) == 1:
        return values[0]
    else:
        return '{0}({1})'.format(glsl_type_of(value), ', '.join(values))


# Dictionary containing the test vectors.  Each entry in the
# dictionary represents a single overload of a single built-in
# function.  Its key is a Signature tuple, and its value is a list of
# TestVector tuples.
#
# Note: the dictionary is initialized to {} here, but it is filled
# with test vectors by code later in this file.
test_suite = {}


# Implementation
# ==============
#
# The functions below shouldn't be necessary to call from outside this
# file.  They exist solely to populate test_suite with test vectors.

# Functions that simulate GLSL built-in functions (in the cases where
# the GLSL built-in functions have no python or numpy equivalent, or
# in cases where there is a behavioral difference).  These functions
# return None if the behavior of the GLSL built-in is undefined for
# the given set of inputs.
def _multiply(x, y):
    x_type = glsl_type_of(x)
    y_type = glsl_type_of(y)

    if x_type.is_vector and y_type.is_vector:
        # vector * vector is done componentwise.
        return x * y
    else:
        # All other cases are standard linear algebraic
        # multiplication, which numpy calls "dot".
        return np.dot(x, y)


def _divide(x, y):
    if any(y_element == 0 for y_element in column_major_values(y)):
        # Division by zero is undefined.
        return None
    return x / y


def _modulus(x, y):
    if any(x_element < 0 for x_element in column_major_values(x)):
        # Modulus operation with a negative first operand is
        # undefined.
        return None
    if any(y_element <= 0 for y_element in column_major_values(y)):
        # Modulus operation with a negative or zero second operand is
        # undefined.
        return None
    return x % y


def _lshift(x, y):
    if not all(0 <= y_element < 32 for y_element in column_major_values(y)):
        # Shifts by less than 0 or more than the number of bits in the
        # type being shifted are undefined.
        return None
    # When the arguments to << don't have the same signedness, numpy
    # likes to promote them to int64.  To avoid this, convert y to be
    # the same type as x.
    y_orig = y
    result = x << y

    # Shifting should always produce a result with the same base type
    # as the left argument.
    assert glsl_type_of(result).base_type == glsl_type_of(x).base_type

    return result


def _rshift(x, y):
    if not all(0 <= y_element < 32 for y_element in column_major_values(y)):
        # Shifts by less than 0 or more than the number of bits in the
        # type being shifted are undefined.
        return None
    # When the arguments to >> don't have the same signedness, numpy
    # likes to promote them to int64.  To avoid this, convert y to be
    # the same type as x.
    y_orig = y
    result = x >> y

    # Shifting should always produce a result with the same base type
    # as the left argument.
    assert glsl_type_of(result).base_type == glsl_type_of(x).base_type

    return result


def _equal(x, y):
    return all(column_major_values(x == y))


def _not_equal(x, y):
    return not _equal(x, y)


def _arctan2(y, x):
    if x == y == 0.0:
        return None
    return np.arctan2(y, x)


def _pow(x, y):
    if x < 0.0:
        return None
    if x == 0.0 and y <= 0.0:
        return None
    return np.power(x, y)


def _exp2(x):
    # exp2() is not available in versions of numpy < 1.3.0 so we
    # emulate it with power().
    return np.power(2, x)


def _trunc(x):
    # trunc() rounds toward zero.  It is not available in version
    # 1.2.1 of numpy so we emulate it with floor(), sign(), and abs().
    return np.sign(x) * np.floor(np.abs(x))


def _clamp(x, minVal, maxVal):
    if minVal > maxVal:
        return None
    return min(max(x, minVal), maxVal)


# Inefficient, but obvious
def _mid3(x, y, z):
    return np.sort([x, y, z])[1]

def _smoothstep(edge0, edge1, x):
    if edge0 >= edge1:
        return None
    t = _clamp((x-edge0)/(edge1-edge0), 0.0, 1.0)
    return t*t*(3.0-2.0*t)


def _normalize(x):
    return x/np.linalg.norm(x)


def _faceforward(N, I, Nref):
    if np.dot(Nref, I) < 0.0:
        return N
    else:
        return -N


def _reflect(I, N):
    return I-2*np.dot(N, I)*N


def _refract(I, N, eta):
    k = 1.0-eta*eta*(1.0-np.dot(N, I)*np.dot(N, I))
    if k < 0.0:
        return I*0.0
    else:
        return eta*I-(eta*np.dot(N, I)+np.sqrt(k))*N


def _argument_types_match(arguments, argument_indices_to_match):
    """Return True if all of the arguments indexed by
    argument_indices_to_match have the same GLSL type.
    """
    types = [glsl_type_of(arguments[i]) for i in argument_indices_to_match]
    return all(x == types[0] for x in types)


def _strict_tolerance(arguments, result):
    """Compute tolerance using a strict interpretation of the GLSL and
    OpenGL standards.

    From the GLSL 1.20 spec (4.1.4 "Floats"):

      "As an input value to one of the processing units, a
      floating-point variable is expected to match the IEEE single
      precision floating-point definition for precision and dynamic
      range.  It is not required that the precision of internal
      processing match the IEEE floating-point specification for
      floating-point operations, but the guidelines for precision
      established by the OpenGL 1.4 specification must be met."

    From the OpenGL 1.4 spec (2.1.1 "Floating-Point Computation"):

      "We require simply that numbers' floating-point parts contain
      enough bits ... so that individual results of floating-point
      operations are accurate to about 1 part in 10^5."

    A harsh interpretation of the above is that (a) no precision is
    lost in moving numbers into or out of the GPU, and (b) any
    built-in function constitutes a single operation, so therefore the
    error in applying any built-in function should be off by no more
    than 1e-5 times its theoretically correct value.

    This is not the only possible interpretation, however.  Certain
    built-in functions, such as the cross product, are computed by a
    formula consisting of many elementary multiplications and
    additions, in which a large amount of cancellation sometimes
    occurs.  It's possible that these rules are meant to apply to
    those elementary multiplications and additions, and not the full
    built-in function. Other built-in functions, such as the trig
    functions, are typically implemented by a series approximation, in
    which 1 part in 10^5 accuracy seems like overkill.  See below for
    the tolerance computation we use on these other functions.
    """
    return 1e-5 * np.linalg.norm(result)


def _trig_tolerance(arguments, result):
    """Compute a more lenient tolerance bound for trig functions.

    The GLSL and OpenGL specs don't provide any guidance as to the
    required accuracy of trig functions (other than the "1 part in
    10^5" general accuracy requirement, which seems like overkill for
    trig functions.

    So the tolerance here is rather arbitrarily chosen to be either 1
    part in 10^3 or 10^-4, whichever is larger.
    """
    return max(1e-4, 1e-3 * np.linalg.norm(result))


def _cross_product_tolerance(arguments, result):
    """Compute a more lenient tolerance bound for cross product.

    Since the computation of a cross product may involve a large
    amount of cancellation, an error tolerance of 1 part in 10^5
    (referred to the magnitude of the result vector) is overly tight.

    So instead we allow the error to be 1 part in 10^5 referred to the
    product of the magnitudes of the arguments.
    """
    assert len(arguments) == 2
    return 1e-5 * np.linalg.norm(arguments[0]) * np.linalg.norm(arguments[1])


def _simulate_function(test_inputs, python_equivalent, tolerance_function):
    """Construct test vectors by simulating a GLSL function on a list
    of possible inputs, and return a list of test vectors.

    test_inputs is a list of possible input sequences, each of which
    represents a set of arguments that should be applied to the
    function.

    python_equivalent is the function to simulate--it should return
    None if the GLSL function returns undefined results for the given
    set of inputs, otherwise it should return the expected result.
    Input sequences for which python_equivalent returns None are
    ignored.

    tolerance_function is the function to call to compute the
    tolerance.  It should take the set of arguments and the expected
    result as its parameters.  It is only used for functions that
    return floating point values.
    """
    test_vectors = []
    for inputs in test_inputs:
        expected_output = python_equivalent(*inputs)
        if expected_output is not None:
            tolerance = tolerance_function(inputs, expected_output)
            test_vectors.append(TestVector(inputs, expected_output, tolerance))
    return test_vectors


def _vectorize_test_vectors(test_vectors, scalar_arg_indices, vector_length):
    """Build a new set of test vectors by combining elements of
    test_vectors into vectors of length vector_length. For example,
    vectorizing the test vectors

    [TestVector((10, 20), 30, tolerance), TestVector((11, 20), 31, tolerance)]

    into vectors of length 2 would produce the result:

    [TestVector((vec2(10, 11), vec2(20, 20)), vec2(30, 31), new_tolerance)].

    Tolerances are combined in root-sum-square fashion.

    scalar_arg_indices is a sequence of argument indices which should
    not be vectorized.  So, if scalar_arg_indices is [1] in the above
    example, the result would be:

    [TestVector((vec2(10, 11), 20), vec2(30, 31), new_tolerance)].
    """
    def make_groups(test_vectors):
        """Group test vectors according to the values passed to the
        arguments that should not be vectorized.
        """
        groups = {}
        for tv in test_vectors:
            key = tuple(tv.arguments[i] for i in scalar_arg_indices)
            if key not in groups:
                groups[key] = []
            groups[key].append(tv)
        return groups

    def partition_vectors(test_vectors, partition_size):
        """Partition test_vectors into lists of length partition_size.
        If partition_size does not evenly divide the number of test
        vectors, wrap around as necessary to ensure that every input
        test vector is included.
        """
        for i in range(0, len(test_vectors), partition_size):
            partition = []
            for j in range(partition_size):
                partition.append(test_vectors[(i + j) % len(test_vectors)])
            yield partition

    def merge_vectors(test_vectors):
        """Merge the given set of test vectors (whose arguments and
        result are scalars) into a single test vector whose arguments
        and result are vectors.  For argument indices in
        scalar_arg_indices, leave the argument as a scalar.
        """
        arity = len(test_vectors[0].arguments)
        arguments = []
        for j in range(arity):
            if j in scalar_arg_indices:
                arguments.append(test_vectors[0].arguments[j])
            else:
                arguments.append(
                    np.array([tv.arguments[j] for tv in test_vectors]))
        result = np.array([tv.result for tv in test_vectors])
        tolerance = np.linalg.norm([tv.tolerance for tv in test_vectors])
        return TestVector(arguments, result, tolerance)

    vectorized_test_vectors = []
    groups = make_groups(test_vectors)
    for key in sorted(groups.keys()):
        test_vectors = groups[key]
        vectorized_test_vectors.extend(
            merge_vectors(partition)
            for partition in partition_vectors(test_vectors, vector_length))
    return vectorized_test_vectors


def _store_test_vector(test_suite_dict, name, glsl_version, extension, test_vector,
                       template=None):
    """Store a test vector in the appropriate place in
    test_suite_dict.  The dictionary key (which is a Signature tuple)
    is generated by consulting the argument and return types of the
    test vector, and combining them with name and glsl_version.

    glsl_version is adjusted if necessary to reflect when the argument
    and return types were introduced into GLSL.

    If template is supplied, it is used insted as the template for the
    Signature objects generated.
    """
    if template is None:
        arg_indices = range(len(test_vector.arguments))
        template = '{0}({1})'.format(
            name, ', '.join('{{{0}}}'.format(i) for i in arg_indices))
    rettype = glsl_type_of(test_vector.result)
    argtypes = tuple(glsl_type_of(arg) for arg in test_vector.arguments)
    adjusted_glsl_version = glsl_version

    signature = Signature(
        name, template, adjusted_glsl_version, extension, rettype, argtypes)
    if signature not in test_suite_dict:
        test_suite_dict[signature] = []
    test_suite_dict[signature].append(test_vector)


def _store_test_vectors(test_suite_dict, name, glsl_version, extension,
                        test_vectors, template=None):
    """Store multiple test vectors in the appropriate places in
    test_suite_dict.

    If template is supplied, it is used insted as the template for the
    Signature objects generated.
    """
    for test_vector in test_vectors:
        _store_test_vector(test_suite_dict, name, glsl_version, extension,
                           test_vector, template=template)


def make_arguments(input_generators):
    """Construct a list of tuples of input arguments to test.

    input_generators is a list, the ith element of which is a sequence
    of values that are suitable for use as the ith argument of the
    function under test.

    Output is a list, each element of which is a tuple of arguments to
    be passed to the function under test.  These values are produced
    by taking the cartesian product of the input sequences.

    """
    return list(itertools.product(*input_generators))


def _make_componentwise_test_vectors(test_suite_dict):
    """Add test vectors to test_suite_dict for GLSL built-in
    functions that operate on vectors in componentwise fashion.
    Examples include sin(), cos(), min(), max(), and clamp().
    """
    # Make sure atan(x) and atan(x,y) don't misbehave for very large
    # or very small input values.
    atan_inputs = [0.0]
    for exponent in (-10, -1, 0, 1, 10):
        atan_inputs.append(pow(10.0, exponent))
        atan_inputs.append(-pow(10.0, exponent))
    # Make a similar set of inputs for acosh(), except don't use any
    # values < 1, since acosh() is only defined for x >= 1.
    acosh_inputs = [1.0 + x for x in atan_inputs if x >= 0]
    ints = [np.int32(x) for x in [-5, -2, -1, 0, 1, 2, 5]]
    uints = [np.uint32(x) for x in [0, 1, 2, 5, 34]]
    bools = [True, False]

    def f(name, arity, python_equivalent,
          alternate_scalar_arg_indices, test_inputs,
          tolerance_function=_strict_tolerance):

        """Create test vectors for the function with the given name
        and arity, which was introduced in the given glsl_version.

        python_equivalent is a Python function which operates on scalars,
        and simulates the GLSL function.  This function should return None
        in any case where the output of the GLSL function is undefined.

        If alternate_scalar_arg_indices is not None, also create test
        vectors for an alternate vectorized version of the function,
        in which some arguments are scalars.
        alternate_scalar_arg_indices is a sequence of the indices of
        the arguments which are scalars.

        test_inputs is a list, the ith element of which is a list of
        values that are suitable for use as the ith argument of the
        function.

        If tolerance_function is supplied, it is a function which
        should be used to compute the tolerance for the test vectors.
        Otherwise, _strict_tolerance is used.
        """
        scalar_test_vectors = _simulate_function(
            make_arguments(test_inputs), python_equivalent, tolerance_function)
        _store_test_vectors(
            test_suite_dict, name, 400, None, scalar_test_vectors)
        _store_test_vectors(
            test_suite_dict, name, 150, "ARB_gpu_shader_fp64", scalar_test_vectors)
        if alternate_scalar_arg_indices is None:
            scalar_arg_indices_list = [()]
        else:
            scalar_arg_indices_list = [(), alternate_scalar_arg_indices]
        for scalar_arg_indices in scalar_arg_indices_list:
            for vector_length in (2, 3, 4):
                _store_test_vectors(
                    test_suite_dict, name, 400, None,
                    _vectorize_test_vectors(
                        scalar_test_vectors, scalar_arg_indices,
                        vector_length))
                _store_test_vectors(
                    test_suite_dict, name, 150, "ARB_gpu_shader_fp64",
                    _vectorize_test_vectors(
                        scalar_test_vectors, scalar_arg_indices,
                        vector_length))

    f('sqrt', 1, np.sqrt, None, [np.linspace(0.0, 2.0, 4)])
    f('inversesqrt', 1, lambda x: 1.0/np.sqrt(x), None,
      [np.linspace(0.1, 2.0, 4)])
    f('abs', 1, np.abs, None, [np.linspace(-1.5, 1.5, 5)])
    f('sign', 1, np.sign, None, [np.linspace(-1.5, 1.5, 5)])
    f('floor', 1, np.floor, None, [np.linspace(-2.0, 2.0, 4)])

    # Note: with trunc we want to test values in which the floating
    # point exponent is < 0, > 52 or in the middle. Hence, we append
    # some numbers to cover all possible scenarios. In addition, we
    # want to check bitsize barriers (> 32, > 64, etc.) in case the
    # conversion is done with a cast to and from another int based
    # type.
    f('trunc', 1, _trunc, None,
      [np.append(np.linspace(-2.0, 2.0, 8),
                 [0.0, 45027112.98414, -45027112.98414,
                  19584198736.47329, -19584198736.47329,
                  7.51654162319e17, -7.51654162319e17,
                  4.502711234095857e19, -4.5027112340958570e19,
                  8.654196813385455e39, -8.6541968133854550e39,
                  6.54965168513655e83, -6.54965168513655e83])])

    # Note: the direction of rounding used by round() is not specified
    # for half-integer values, so we test it over a range that doesn't
    # include exact half-integer values.  roundEven() is required to
    # round half-integer values to the nearest even integer, so we
    # test it over a range that does include exact half-integer
    # values.  In both cases, we can use numpy's round() function,
    # because it rounds half-integer values to even, and all other
    # values to nearest.
    f('round', 1, np.round, None, [np.linspace(-2.0, 2.0, 8)])
    f('roundEven', 1, np.round, None, [np.linspace(-2.0, 2.0, 25)])

    f('ceil', 1, np.ceil, None, [np.linspace(-2.0, 2.0, 4)])
    f('fract', 1, lambda x: x-np.floor(x), None,
      [np.linspace(-2.0, 2.0, 4)])
    f('mod', 2, lambda x, y: x-y*np.floor(x/y), [1],
      [np.linspace(-1.9, 1.9, 4), np.linspace(-2.0, 2.0, 4)])
    f('min', 2, min, [1],
      [np.linspace(-2.0, 2.0, 4), np.linspace(-2.0, 2.0, 4)])
    f('max', 2, max, [1],
      [np.linspace(-2.0, 2.0, 4), np.linspace(-2.0, 2.0, 4)])
    f('clamp', 3, _clamp, [1, 2], [np.linspace(-2.0, 2.0, 4),
      np.linspace(-1.5, 1.5, 3), np.linspace(-1.5, 1.5, 3)])
    f('mix', 3, lambda x, y, a: x*(1-a)+y*a, [2],
      [np.linspace(-2.0, 2.0, 2), np.linspace(-3.0, 3.0, 2),
       np.linspace(0.0, 1.0, 4)])
    f('mix', 3, lambda x, y, a: y if a else x, None,
      [np.linspace(-2.0, 2.0, 2), np.linspace(-3.0, 3.0, 2), bools])
    f('step', 2, lambda edge, x: 0.0 if x < edge else 1.0, [0],
      [np.linspace(-2.0, 2.0, 4), np.linspace(-2.0, 2.0, 4)])
    f('smoothstep', 3, _smoothstep, [0, 1],
      [np.linspace(-1.9, 1.9, 4), np.linspace(-1.9, 1.9, 4),
       np.linspace(-2.0, 2.0, 4)])

_make_componentwise_test_vectors(test_suite)


def _make_vector_relational_test_vectors(test_suite_dict):
    """Add test vectors to test_suite_dict for GLSL built-in functions
    that operate on vectors of floats, ints, or bools, but not on
    single floats, ints, or bools.  Examples include lessThan(),
    equal(), and not().
    """
    _default_inputs = {
        'v': np.linspace(-1.5, 1.5, 4),
        'b': np.array([False, True])
        }

    def f(name, arity, python_equivalent, arg_types,
          tolerance_function=_strict_tolerance,
          extension=None):
        """Make test vectors for the function with the given name and
        arity, which was introduced in the given glsl_version.

        python_equivalent is a Python function which operates on scalars,
        and simulates the GLSL function.

        arg_types is a string containing 'v' if the function supports
        standard "vec" inputs, 'i' if it supports "ivec" inputs, and 'b'
        if it supports "bvec" inputs.  The output type of the function is
        assumed to be the same as its input type.

        If tolerance_function is supplied, it is a function which
        should be used to compute the tolerance for the test vectors.
        Otherwise, _strict_tolerance is used.
        """
        for arg_type in arg_types:
            test_inputs = [_default_inputs[arg_type]]*arity
            scalar_test_vectors = _simulate_function(
                make_arguments(test_inputs), python_equivalent,
                tolerance_function)
            for vector_length in (2, 3, 4):
                _store_test_vectors(
                    test_suite_dict, name, 400, None,
                    _vectorize_test_vectors(
                        scalar_test_vectors, (), vector_length))
                _store_test_vectors(
                    test_suite_dict, name, 150, "ARB_gpu_shader_fp64",
                    _vectorize_test_vectors(
                        scalar_test_vectors, (), vector_length))

    f('lessThan', 2, lambda x, y: x < y, 'v')
    f('lessThanEqual', 2, lambda x, y: x <= y, 'v')
    f('greaterThan', 2, lambda x, y: x > y, 'v')
    f('greaterThanEqual', 2, lambda x, y: x >= y, 'v')
    f('equal', 2, lambda x, y: x == y, 'v')
    f('notEqual', 2, lambda x, y: x != y, 'v')

_make_vector_relational_test_vectors(test_suite)


def _make_vector_or_matrix_test_vectors(test_suite_dict):
    """Add test vectors to test_suite_dict for GLSL built-in functions
    that operate on vectors/matrices as a whole.  Examples include
    length(), dot(), cross(), normalize(), and refract().
    """
    def match_args(*indices):
        """Return a function that determines whether the type of the
        arguments at the given indices match.

        For example:

            match(1, 3)

        is equivalent to:

            lambda a, b, c, d: glsl_type_of(b) == glsl_type_of(d)
        """
        return lambda *args: _argument_types_match(args, indices)

    def match_simple_binop(x, y):
        """Detemine whether the type of the arguments is compatible
        for a simple binary operator (such as '+').

        Arguments are compatible if one is a scalar and the other is a
        vector/matrix with the same base type, or if they are the same
        type.
        """
        x_type = glsl_type_of(x)
        y_type = glsl_type_of(y)
        if x_type.base_type != y_type.base_type:
            return False
        if x_type.is_scalar or y_type.is_scalar:
            return True
        return x_type == y_type

    def match_multiply(x, y):
        """Determine whether the type of the arguments is compatible
        for multiply.

        Arguments are compatible if they are scalars, vectors, or
        matrices with the same base type, and the vector/matrix sizes
        are properly matched.
        """
        x_type = glsl_type_of(x)
        y_type = glsl_type_of(y)
        if x_type.base_type != y_type.base_type:
            return False
        if x_type.is_scalar or y_type.is_scalar:
            return True
        if x_type.is_vector and y_type.is_matrix:
            # When multiplying vector * matrix, the vector is
            # transposed to a row vector.  So its row count must match
            # the row count of the matrix.
            return x_type.num_rows == y_type.num_rows
        elif x_type.is_vector:
            assert y_type.is_vector
            # When multiplying vector * vector, the multiplication is
            # done componentwise, so the types must match exactly.
            return x_type == y_type
        else:
            assert x_type.is_matrix
            # When multiplying matrix * matrix or matrix * vector, a
            # standard linear algebraic multiply is used, so x's
            # column count must match y's row count.
            return x_type.num_cols == y_type.num_rows

    def match_shift(x, y):
        """Determine whether the type of the arguments is compatible
        for shift operations.

        Arguments are compatible if they are the same length or the
        first one is a vector and the second is a scalar.  Their base
        types need not be the same, but they both must be integral.
        """
        x_type = glsl_type_of(x)
        y_type = glsl_type_of(y)
        if x_type.base_type not in (glsl_int, glsl_uint):
            return False
        if y_type.base_type not in (glsl_int, glsl_uint):
            return False
        if y_type.is_scalar:
            return True
        assert not x_type.is_matrix
        assert not y_type.is_matrix
        return x_type.num_rows == y_type.num_rows

    nz_doubles = [ -1.333333333333333259, 0.85]
    doubles = [0.0] + nz_doubles
    dvecs = [
        np.array([-0.10, -1.20]),
        np.array([-0.42, 0.48]),
        np.array([-1.333333333333333259, -0.85, -0.94]),
        np.array([1.67, 0.66, 1.87]),
        np.array([-1.65, 1.33, 1.93, 0.76]),
        np.array([0.80, -0.15, -0.51, 0.0])
        ]
    nz_doubles_dvecs = nz_doubles + dvecs
    dvec3s = [
        np.array([-0.03, -0.85, -0.94]),
        np.array([ -1.333333333333333259, 0.66, 1.87]),
        ]

    norm_doubles_dvecs = [_normalize(x) for x in nz_doubles_dvecs]
    squaremats = [
        np.array([[ 1.60,  0.76],
                  [ 1.53, -1.00]]),  # mat2
        np.array([[-0.13, -0.87],
                  [-1.40,  1.40]]),  # mat2
        np.array([[-1.11,  1.67, -0.41],
                  [ 0.13,  1.09, -0.02],
                  [ 0.56,  0.95,  0.24]]),  # mat3
        np.array([[-1.69, -0.46, -0.18],
                  [-1.09,  1.75,  2.00],
                  [-1.53, -0.70, -1.47]]),  # mat3
        np.array([[-1.00, -0.55, -1.08,  1.79],
                  [ 1.77,  0.62,  0.48, -1.35],
                  [ 0.09, -0.71, -1.39, -1.21],
                  [-0.91, -1.82, -1.43,  0.72]]),  # mat4
        np.array([[ 0.06,  1.31,  1.52, -1.96],
                  [ 1.60, -0.32,  0.51, -1.84],
                  [ 1.25,  0.45,  1.90, -0.72],
                  [-0.16,  0.45, -0.88,  0.39]]),  # mat4
        ]
    mats = squaremats + [
        np.array([[ 0.09,  1.30,  1.25],
                  [-1.19,  0.08,  1.08]]),  # mat3x2
        np.array([[-0.36, -1.08, -0.60],
                  [-0.53,  0.88, -1.79]]),  # mat3x2
        np.array([[-0.46,  1.94],
                  [-0.45, -0.75],
                  [ 1.03, -0.50]]),  # mat2x3
        np.array([[ 1.38, -1.08],
                  [-1.27,  1.83],
                  [ 1.00, -0.74]]),  # mat2x3
        np.array([[ 1.81, -0.87,  0.81,  0.65],
                  [-1.16, -1.52,  0.25, -1.51]]),  # mat4x2
        np.array([[ 1.93, -1.63,  0.29,  1.60],
                  [ 0.49,  0.27,  0.14,  0.94]]),  # mat4x2
        np.array([[ 0.16, -1.69],
                  [-0.80,  0.59],
                  [-1.74, -1.43],
                  [-0.02, -1.21]]),  # mat2x4
        np.array([[-1.02,  0.74],
                  [-1.64, -0.13],
                  [-1.59,  0.47],
                  [ 0.30,  1.13]]),  # mat2x4
        np.array([[-0.27, -1.38, -1.41, -0.12],
                  [-0.17, -0.56,  1.47,  1.86],
                  [-1.85, -1.29,  1.77,  0.01]]),  # mat4x3
        np.array([[-0.47, -0.15,  1.97, -1.05],
                  [-0.20,  0.53, -1.82, -1.41],
                  [-1.39, -0.19,  1.62,  1.58]]),  # mat4x3
        np.array([[ 1.42, -0.86,  0.27],
                  [ 1.80, -1.74,  0.04],
                  [-1.88, -0.37,  0.43],
                  [ 1.37,  1.90,  0.71]]),  # mat3x4
        np.array([[-1.72,  0.09,  0.45],
                  [-0.31, -1.58,  1.92],
                  [ 0.14,  0.18, -0.56],
                  [ 0.40, -0.77,  1.76]]),  # mat3x4
        ]

    dsquaredmats = [
        np.array([[ 1.60,  0.76],
                  [ 1.53, -1.00]]),  # mat2
        np.array([[-0.13, -0.87],
                  [-1.40,  1.40]]),  # mat2
        np.array([[-1.11,  1.67, -0.41],
                  [ 0.13,  1.09, -0.02],
                  [ 0.56,  0.95,  0.24]]),  # mat3
        np.array([[-1.69, -0.46, -0.18],
                  [-1.09,  1.75,  2.00],
                  [-1.53, -0.70, -1.47]]),  # mat3
        np.array([[-1.00, -0.55, -1.08,  1.79],
                  [ 1.77,  0.62,  0.48, -1.35],
                  [ 0.09, -0.71, -1.39, -1.21],
                  [-0.91, -1.82, -1.43,  0.72]]),  # mat4
        np.array([[ 0.06,  1.31,  1.52, -1.96],
                  [ 1.60, -0.32,  0.51, -1.84],
                  [ 1.25,  0.45,  1.90, -0.72],
                  [-0.16,  0.45, -0.88,  0.39]]),  # mat4
        ]
    dmats = dsquaredmats + [
        np.array([[ 0.09,  1.30,  1.25],
                  [-1.19,  0.08,  1.08]]),  # mat3x2
        np.array([[-0.36, -1.08, -0.60],
                  [-0.53,  0.88, -1.79]]),  # mat3x2
        np.array([[-0.46,  1.94],
                  [-0.45, -0.75],
                  [ 1.03, -0.50]]),  # mat2x3
        np.array([[ 1.38, -1.08],
                  [-1.27,  1.83],
                  [ 1.00, -0.74]]),  # mat2x3
        np.array([[ 1.81, -0.87,  0.81,  0.65],
                  [-1.16, -1.52,  0.25, -1.51]]),  # mat4x2
        np.array([[ 1.93, -1.63,  0.29,  1.60],
                  [ 0.49,  0.27,  0.14,  0.94]]),  # mat4x2
        np.array([[ 0.16, -1.69],
                  [-0.80,  0.59],
                  [-1.74, -1.43],
                  [-0.02, -1.21]]),  # mat2x4
        np.array([[-1.02,  0.74],
                  [-1.64, -0.13],
                  [-1.59,  0.47],
                  [ 0.30,  1.13]]),  # mat2x4
        np.array([[-0.27, -1.38, -1.41, -0.12],
                  [-0.17, -0.56,  1.47,  1.86],
                  [-1.85, -1.29,  1.77,  0.01]]),  # mat4x3
        np.array([[-0.47, -0.15,  1.97, -1.05],
                  [-0.20,  0.53, -1.82, -1.41],
                  [-1.39, -0.19,  1.62,  1.58]]),  # mat4x3
        np.array([[ 1.42, -0.86,  0.27],
                  [ 1.80, -1.74,  0.04],
                  [-1.88, -0.37,  0.43],
                  [ 1.37,  1.90,  0.71]]),  # mat3x4
        np.array([[-1.72,  0.09,  0.45],
                  [-0.31, -1.58,  1.92],
                  [ 0.14,  0.18, -0.56],
                  [ 0.40, -0.77,  1.76]]),  # mat3x4
        ]
    def f(name, arity, python_equivalent,
          filter, test_inputs, tolerance_function=_strict_tolerance,
          template=None):
        """Make test vectors for the function with the given name and
        arity, which was introduced in the given glsl_version.

        python_equivalent is a Python function which simulates the GLSL
        function.  This function should return None in any case where the
        output of the GLSL function is undefined.  However, it need not
        check that the lengths of the input vectors are all the same.

        If filter is not None, it will be called with each set of
        arguments, and test cases will only be generated if the filter
        returns True.

        test_inputs is a list, the ith element of which is a list of
        vectors and/or scalars that are suitable for use as the ith
        argument of the function.

        If tolerance_function is supplied, it is a function which
        should be used to compute the tolerance for the test vectors.
        Otherwise, _strict_tolerance is used.

        If template is supplied, it is used insted as the template for
        the Signature objects generated.
        """
        test_inputs = make_arguments(test_inputs)

        if filter is not None:
            test_inputs = \
                [arguments for arguments in test_inputs if filter(*arguments)]
        _store_test_vectors(
            test_suite_dict, name, 400, None,
            _simulate_function(
                test_inputs, python_equivalent, tolerance_function),
            template=template)
        _store_test_vectors(
            test_suite_dict, name, 150, "ARB_gpu_shader_fp64",
            _simulate_function(
                test_inputs, python_equivalent, tolerance_function),
            template=template)

    f('op-add', 2, lambda x, y: x + y, match_simple_binop,
      [doubles+dvecs+dmats,
       doubles+dvecs+dmats],
      template='({0} + {1})')
    f('op-sub', 2, lambda x, y: x - y, match_simple_binop,
      [doubles+dvecs+dmats,
       doubles+dvecs+dmats],
      template='({0} - {1})')
    f('op-mult', 2, _multiply, match_multiply,
      [doubles+dvecs+dmats,
       doubles+dvecs+dmats],
      template='({0} * {1})')
    f('op-div', 2, _divide, match_simple_binop,
      [doubles+dvecs+dmats,
       doubles+dvecs+dmats],
      template='({0} / {1})')
    f('length', 1, np.linalg.norm, None, [doubles+dvecs])
    f('distance', 2, lambda x, y: np.linalg.norm(x-y), match_args(0, 1),
      [doubles+dvecs, doubles+dvecs])
    f('dot', 2, np.dot, match_args(0, 1), [doubles+dvecs, doubles+dvecs])
    f('cross', 2, np.cross, match_args(0, 1), [dvec3s, dvec3s],
      _cross_product_tolerance)
    f('normalize', 1, _normalize, None, [nz_doubles_dvecs])
    f('faceforward', 3, _faceforward, match_args(0, 1, 2),
      [doubles+dvecs, doubles+dvecs, doubles+dvecs])
    f('reflect', 2, _reflect, match_args(0, 1),
      [doubles+dvecs, norm_doubles_dvecs])
    f('refract', 3, _refract, match_args(0, 1),
      [norm_doubles_dvecs, norm_doubles_dvecs, [0.5, 2.0]])
    f('matrixCompMult', 2, lambda x, y: x*y, match_args(0, 1),
      [dmats, dmats])
    f('outerProduct', 2, np.outer, None, [dvecs, dvecs])
    f('transpose', 1, np.transpose, None, [dmats])

    f('inverse', 1, np.linalg.inv, None, [dsquaredmats])

    f('determinant', 1, np.linalg.det, None, [dsquaredmats])
_make_vector_or_matrix_test_vectors(test_suite)


def _check_signature_safety(test_suite_dict):
    """As a final safety check, verify that for each possible
    combination of name and argtypes, there is exactly one
    signature.
    """
    name_argtype_combos = set()
    for signature in test_suite_dict:
        name_argtype_combo = (signature.name, signature.argtypes, signature.extension)
        if name_argtype_combo in name_argtype_combos:
            raise Exception(
                'Duplicate signature found for {0}'.format(name_argtype_combo))
        name_argtype_combos.add(name_argtype_combo)
_check_signature_safety(test_suite)



# -*- coding: utf-8 -*-

from __future__ import (
     absolute_import, division, print_function, unicode_literals
)
import itertools

import six

from tests.quick_gl import profile as _profile1
from tests.quick_shader import profile as _profile2
from framework.test import PiglitGLTest
from framework.test.shader_test import ShaderTest, MultiShaderTest
from framework.profile import TestProfile

__all__ = ['profile']

# Save the filters from the original profiles to a new profile
profile = TestProfile()
profile.filters = _profile1.filters + _profile2.filters

# Add a modified version of each PiglitGLTest as a khr_no_error variant.
# Shader runner doesn't explitly test for expected errors so we add shader
# tests as is. We actively filter GLSL parser and any other type of tests.
for name, test in itertools.chain(six.iteritems(_profile1.test_list),
                                  six.iteritems(_profile2.test_list)):
    if isinstance(test, (PiglitGLTest, ShaderTest, MultiShaderTest)):
        profile.test_list['{} khr_no_error'.format(name)] = test
        test.command += ['-khr_no_error']



# coding=utf-8
# Copyright (c) 2016 Broadcom
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice (including the next
# paragraph) shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

"""Tests for the driver_classifier module."""

from __future__ import (
    absolute_import, division, print_function, unicode_literals
)

try:
    import mock
except ImportError:
    from unittest import mock

import pytest
import six

from framework import driver_classifier


class DriverClassifierTester(driver_classifier.DriverClassifier):
    """Test class for the driver classifier, taking in a fixed

    renderer string instead of calling glxinfo.
    """

    def __init__(self, renderer):
        self.override_renderer = renderer
        super(DriverClassifierTester, self).__init__()


    def collect_driver_info(self):
        self.renderer = self.override_renderer


class TestDriverClassifier(object):
    """Tests for the DriverClassifier class."""

    @pytest.mark.parametrize(
        "renderer, categories",
        [
            ('Mesa DRI Intel(R) Haswell Mobile', ['i965-hsw', 'i965', 'mesa']),
            ('Mesa DRI Intel(R) HD Graphics 530 (Skylake GT2)', ['i965-skl', 'i965', 'mesa']),
            # Old VC4 string
            ('Gallium 0.4 on VC4', ['vc4', 'mesa']),
            # Modern VC4 string
            ('Gallium 0.4 on VC4 V3D 2.1', ['vc4-2.1', 'vc4', 'mesa']),
        ],
        ids=six.text_type)
    def test_renderer_categorization(self, renderer, categories):
        """Test that when given a certain renderer string, the correct

        categories list comes back.
        """
        assert DriverClassifierTester(renderer).categories == categories

    def test_collect_glxinfo(self):
        """Should set self.renderer."""
        test = driver_classifier.DriverClassifier()
        with mock.patch('framework.driver_classifier.subprocess.check_output',
                        mock.Mock(return_value=b'some data\nand some more\n'
                                               b'OpenGL renderer string: '
                                               b'sentinal\nand some other '
                                               b'stuff')):
            test.collect_glxinfo()
        assert isinstance(test.renderer, six.text_type)
        assert test.renderer == 'sentinal'



# encoding=utf-8
# Copyright © 2018 Intel Corporation

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""A minimal profile that aims to touch test most major OpenGL features with as
few tests as possible.

This is largely aimed at bring up of a new driver or a driver on new hardware,
where performance and stability are not necessarily the best. The idea is to
touch the main functionality of important features without running for too
long. This *will* miss many corner cases, that's okay. For more stable
hardware or software stacks the "gpu" or "quick" profiles are probably more
appropriate.
"""

from __future__ import (
    absolute_import, division, print_function, unicode_literals
)
import os

from framework import grouptools
from framework.profile import TestProfile
from framework.test import PiglitGLTest
from framework.test.shader_test import ShaderTest
from .py_modules.constants import GENERATED_TESTS_DIR, TESTS_DIR

__all__ = ['profile']

basepath = os.path.normpath(os.path.join(TESTS_DIR, '..'))
gen_basepath = os.path.relpath(os.path.join(GENERATED_TESTS_DIR, '..'), basepath)


def add_shader_test(shader):
    """Given an adder, creates a group and adds a shader test."""
    for d in (TESTS_DIR, GENERATED_TESTS_DIR,):
        s = os.path.abspath(os.path.join(d, shader))
        if os.path.exists(s):
            basedir = d
            abs_path = s
            break

    dirpath, filename = os.path.split(os.path.join(basedir, shader))
    dirname = os.path.relpath(dirpath, basepath)
    filepath = os.path.join(dirname, filename)

    if os.path.commonprefix([abs_path, GENERATED_TESTS_DIR]) == GENERATED_TESTS_DIR:
        installpath = os.path.relpath(filepath, gen_basepath)
    else:
        installpath = None

    groupname = grouptools.from_path(os.path.relpath(dirpath, basedir))
    testname = os.path.splitext(os.path.basename(shader))[0]
    group = grouptools.join(groupname, testname)
    profile.test_list[group] = ShaderTest.new(filepath, installpath)


profile = TestProfile()

shader_tests = (
    'fast_color_clear/all-colors.shader_test',
    'shaders/arb_texture_gather-miplevels.shader_test',
    'shaders/glsl-algebraic-add-zero.shader_test',
    'shaders/glsl-derivs-varyings.shader_test',
    'shaders/glsl-fs-max-3.shader_test',
    'shaders/glsl-fs-texture2d-branching.shader_test',
    'shaders/glsl-vs-mat-mul-1.shader_test',
    'spec/amd_shader_trinary_minmax/execution/built-in-functions/cs-max3-uvec4-uvec4-uvec4.shader_test',
    'spec/amd_shader_trinary_minmax/execution/built-in-functions/tcs-max3-vec4-vec4-vec4.shader_test',
    'spec/arb_arrays_of_arrays/execution/image_store/basic-imageStore-mixed-const-non-const-uniform-index.shader_test',
    'spec/arb_arrays_of_arrays/execution/sampler/vs-struct-nonconst.shader_test',
    'spec/arb_compute_shader/execution/basic-ssbo.shader_test',
    'spec/arb_compute_shader/execution/shared-atomics.shader_test',
    'spec/arb_derivative_control/execution/dfdx-dfdy.shader_test',
    'spec/arb_gpu_shader5/execution/built-in-functions/fs-bitfieldReverse.shader_test',
    'spec/arb_gpu_shader_fp64/execution/built-in-functions/fs-distance-dvec4-dvec4.shader_test',
    'spec/arb_gpu_shader_int64/execution/built-in-functions/cs-op-div-i64vec4-i64vec4.shader_test',
    'spec/arb_shader_ballot/execution/fs-readFirstInvocation-uint-loop.shader_test',
    'spec/arb_shader_group_vote/cs-all.shader_test',
    'spec/arb_shader_image_load_store/execution/write-to-rendered-image.shader_test',
    'spec/arb_shader_precision/gs-op-div-vec4-vec4.shader_test',
    'spec/arb_shader_subroutine/execution/two-subroutines-nested.shader_test',
    'spec/arb_shader_texture_lod/execution/glsl-fs-shadow2DGradARB-cumulative.shader_test',
    'spec/arb_tessellation_shader/execution/barrier-patch.shader_test',
    'spec/arb_tessellation_shader/execution/built-in-functions/tcs-any-bvec4-using-if.shader_test',
    'spec/arb_tessellation_shader/execution/sanity.shader_test',
    'spec/arb_vertex_attrib_64bit/execution/vs_in/vs-input-uint_uvec4-double_dmat3x4_array2-position.shader_test',
    'spec/glsl-1.50/execution/geometry-basic.shader_test',
    'spec/oes_viewport_array/viewport-gs-write-simple.shader_test',
)

for shader_test in shader_tests:
    add_shader_test(shader_test)

with profile.test_list.group_manager(
        PiglitGLTest,
        grouptools.join('spec', 'glsl-1.50')) as g:
    g(['glsl-1.50-transform-feedback-builtins'], 'transform-feedback-builtins')



# coding=utf-8
# Copyright (c) 2015-2016 Intel Corporation

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Tests for framework.summary.console_.

Some of these tests are fickle, since we're testing the output of a summary
generator.
"""

from __future__ import (
    absolute_import, division, print_function, unicode_literals
)

import pytest
import six

from framework import results
from framework import grouptools
from framework.summary import console_, common

# pylint: disable=no-self-use,protected-access

_ENUMS = {
    1: 'names',
    2: 'divider',
    3: 'pass',
    4: 'fail',
    5: 'crash',
    6: 'skip',
    7: 'timeout_',  # "timeout" interferes with py.test's dependency injection
    8: 'warn',
    9: 'incomplete',
    10: 'dmesg-warn',
    11: 'dmesg-fail',
    12: 'changes',
    13: 'fixes',
    14: 'regressions',
    15: 'total',
    16: 'time',
}

class TestPrintSummary(object):
    """Tests for the console output."""

    @pytest.mark.parametrize("index", six.iterkeys(_ENUMS), ids=lambda i: _ENUMS[i])
    def test_values(self, index, capsys):
        """Create both an expected value and an actual value.

        The expected value makes use of the template, this helps to minimize
        the number of changes that need to be made if the output is altered.
        """
        names = [grouptools.join('foo', 'bar', 'oink', 'foobar', 'boink'),
                 'foo', 'bar']
        template = '{: >20.20} {: >12.12}'

        expected = console_._SUMMARY_TEMPLATE.format(
            names=' '.join(['this is a really rea', 'another name']),
            divider=' '.join(['--------------------', '------------']),
            pass_=template.format('1', '2'),
            fail=template.format('2', '0'),
            crash=template.format('0', '0'),
            skip=template.format('0', '1'),
            timeout=template.format('0', '0'),
            warn=template.format('0', '0'),
            incomplete=template.format('0', '0'),
            dmesg_warn=template.format('0', '0'),
            dmesg_fail=template.format('0', '0'),
            changes=template.format('0', '2'),
            fixes=template.format('0', '1'),
            regressions=template.format('0', '0'),
            total=template.format('3', '3'),
            time=template.format('00:01:39', '02:14:05')).split('\n')

        res1 = results.TestrunResult()
        res1.name = 'this is a really really really really long name'
        res1.tests[names[0]] = results.TestResult('pass')
        res1.tests[names[1]] = results.TestResult('fail')
        res1.tests[names[2]] = results.TestResult('notrun')
        res1.tests[names[2]].subtests['1'] = 'fail'
        res1.time_elapsed = results.TimeAttribute(1509747121.4873962, 1509747220.544042)
        res1.calculate_group_totals()

        res2 = results.TestrunResult()
        res2.name = 'another name'
        res2.tests[names[0]] = results.TestResult('pass')
        res2.tests[names[1]] = results.TestResult('pass')
        res2.tests[names[2]] = results.TestResult('notrun')
        res2.tests[names[2]].subtests['1'] = 'skip'
        res2.time_elapsed = results.TimeAttribute(1464820707.4581327, 1464828753.201948)
        res2.calculate_group_totals()

        reses = common.Results([res1, res2])

        console_._print_summary(reses)
        actual = capsys.readouterr()[0].splitlines()

        assert actual[index] == expected[index]


class TestPrintResult(object):
    """Tests for the _print_result function."""

    def test_basic(self, capsys):
        """summary.console_._print_result: prints expected values."""
        res1 = results.TestrunResult()
        res1.tests['foo'] = results.TestResult('pass')

        res2 = results.TestrunResult()
        res2.tests['foo'] = results.TestResult('fail')

        reses = common.Results([res1, res2])

        expected = 'foo: pass fail\n'
        console_._print_result(reses, reses.names.all)
        actual, _ = capsys.readouterr()

        assert expected == actual

    def test_replaces_separator(self, capsys):
        """summary.console_._print_result: Replaces separator with /."""
        res1 = results.TestrunResult()
        res1.tests[grouptools.join('foo', 'bar')] = results.TestResult('pass')

        reses = common.Results([res1])

        expected = 'foo/bar: pass\n'
        console_._print_result(reses, reses.names.all)
        actual, _ = capsys.readouterr()

        assert expected == actual



# coding=utf-8
# Copyright (c) 2014, 2016 Intel Corporation

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Tests for log.py module."""

from __future__ import (
    absolute_import, division, print_function, unicode_literals
)
import collections
import sys
import threading

import pytest
import six

import framework.log as log

# pylint: disable=no-self-use,protected-access


@pytest.fixture
def log_state():
    """Create a unique state instance per test."""
    return {'total': 1, 'complete': 0, 'lastlength': 0, 'running': [],
            'summary': collections.defaultdict(lambda: 0)}


class TestLogFactory(object):
    """Tests for the LogFactory class."""

    class TestGet(object):
        """Tests for the LogFactory.get method."""

        def test_returns_log(self):
            """Returns a BaseLog derived instance."""
            logger = log.LogManager('quiet', 100)
            log_inst = logger.get()
            assert isinstance(log_inst, log.BaseLog)

    def test_log_state_update(self):
        """log.BaseLog.log updates shared state managed by LogManager"""
        logger = log.LogManager('quiet', 100)
        log_inst = logger.get()
        log_inst.start(None)
        log_inst.log('pass')

        assert logger._state['total'] == 100
        assert logger._state['summary'] == {'pass': 1}
        assert logger._state['complete'] == 1


class TestQuietLog(object):
    """Test QuietLog class."""

    class TestOutput(object):
        """Test the output of the various methods."""

        @pytest.fixture(autouse=True, scope='function')
        def mock_stdout(self, mocker):
            mocker.patch.object(sys, 'stdout', six.StringIO())

        def test_log(self, log_state):  # pylint: disable=redefined-outer-name
            """Test the output of the log method."""
            quiet = log.QuietLog(log_state, threading.Lock())
            quiet.start(None)
            quiet.log('pass')
            sys.stdout.seek(0)

            actual = sys.stdout.read()
            assert actual == b'[1/1] pass: 1 -\n'

        def test_summary(self, log_state):  # pylint: disable=redefined-outer-name
            """Test the output of the summary method."""
            quiet = log.QuietLog(log_state, threading.Lock())
            # Call log to set the total correctly, then truncate and remove the
            # values, so the we can test
            quiet.start(None)
            quiet.log('pass')
            sys.stdout.seek(0)
            sys.stdout.truncate()

            quiet.summary()
            sys.stdout.seek(0)

            # Because of the 'lastlength' mechanims there will likely be
            # trainling whitespace after the the output, it's not useful to
            # test that here, so just strip it.
            assert sys.stdout.read().rstrip() == b'[1/1] pass: 1'

        def test_start(self, log_state):  # pylint: disable=redefined-outer-name
            """Test that the start method doesn't have output."""
            quiet = log.QuietLog(log_state, threading.Lock())
            quiet.start(None)
            quiet.start('foo')
            sys.stdout.seek(0)

            actual = sys.stdout.read()
            assert actual == b''


class TestVerboseLog(object):
    """Tests for the VerboseLog class."""

    class TestOutput(object):
        """Test the output of the various methods."""

        @pytest.fixture(autouse=True, scope='function')
        def mock_stdout(self, mocker):
            mocker.patch.object(sys, 'stdout', six.StringIO())

        def test_log(self, log_state):  # pylint: disable=redefined-outer-name
            """Test the output of the log method."""
            l = log.VerboseLog(log_state, threading.Lock())
            l.start('foo')
            sys.stdout.seek(0)
            sys.stdout.truncate()

            l.log('pass')
            sys.stdout.seek(0)

            actual = sys.stdout.read()
            assert actual == b'pass: foo\n\n[1/1] pass: 1 /\n'

        def test_summary(self, log_state):  # pylint: disable=redefined-outer-name
            """Test the output of the summary method."""
            l = log.VerboseLog(log_state, threading.Lock())
            l.start('foo')
            l.log('pass')
            sys.stdout.seek(0)
            sys.stdout.truncate()

            l.summary()
            sys.stdout.seek(0)

            assert sys.stdout.read().rstrip() == b'[1/1] pass: 1'

        def test_start(self, log_state):  # pylint: disable=redefined-outer-name
            """Test that the start method doesn't have output."""
            l = log.VerboseLog(log_state, threading.Lock())
            l.start('foo')
            sys.stdout.seek(0)

            assert sys.stdout.read().rstrip() == b'running: foo\n\n[0/1]  \\'


class TestDummyLog(object):
    """Tests for the DummyLog class."""

    class TestOutput(object):
        """Test the output of the various methods."""

        @pytest.fixture(autouse=True, scope='function')
        def mock_stdout(self, mocker):
            mocker.patch.object(sys, 'stdout', six.StringIO())

        def test_log(self, log_state):  # pylint: disable=redefined-outer-name
            """Test the output of the log method."""
            quiet = log.DummyLog(log_state, threading.Lock())
            quiet.start(None)
            quiet.log('pass')
            sys.stdout.seek(0)

            actual = sys.stdout.read()
            assert actual == b''

        def test_summary(self, log_state):  # pylint: disable=redefined-outer-name
            """Test the output of the summary method."""
            quiet = log.DummyLog(log_state, threading.Lock())
            quiet.summary()
            sys.stdout.seek(0)

            actual = sys.stdout.read()
            assert actual == b''

        def test_start(self, log_state):  # pylint: disable=redefined-outer-name
            """Test that the start method doesn't have output."""
            quiet = log.DummyLog(log_state, threading.Lock())
            quiet.start('foo')
            sys.stdout.seek(0)

            actual = sys.stdout.read()
            assert actual == b''



# coding=utf-8
# Copyright (c) 2014, 2016 Intel Corporation

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Tests for the Junit backend package."""

from __future__ import (
    absolute_import, division, print_function, unicode_literals
)
import os
import textwrap
try:
    from lxml import etree
except ImportError:
    import xml.etree.cElementTree as etree
try:
    import mock
except ImportError:
    from unittest import mock

import pytest
import six

from framework import backends
from framework import grouptools
from framework import results
from framework import status

from . import shared

# pylint: disable=no-self-use

JUNIT_SCHEMA = os.path.join(os.path.dirname(__file__), 'schema', 'junit-7.xsd')

_XML = """\
<?xml version='1.0' encoding='utf-8'?>
  <testsuites>
    <testsuite name="piglit" tests="1">
      <testcase classname="piglit.foo.bar" name="a-test" status="pass" time="1.12345">
        <system-out>this/is/a/command\nThis is stdout</system-out>
        <system-err>this is stderr

pid: [1934]
time start: 1.0
time end: 4.5
        </system-err>
      </testcase>
    </testsuite>
  </testsuites>
"""

@pytest.yield_fixture(autouse=True, scope="module")
def mock_compression():
    with mock.patch('framework.backends.compression.get_mode',
                    mock.Mock(return_value='none')):
        yield


class TestProtectedLoad(object):
    """Tests for the _load method."""

    def test_default_name(self, tmpdir):
        """backends.junit._load: uses 'junit result' for name as fallback."""
        tmpdir.chdir()
        with open('results.xml', 'w') as f:
            f.write(_XML)
        test = backends.junit.REGISTRY.load('results.xml', 'none')

        assert test.name == 'junit result'

    def test_file_name(self, tmpdir):
        """backends.junit._load: uses the filename for name if filename !=
        'results'
        """
        p = tmpdir.join('foobar.xml')
        p.write(_XML)
        test = backends.junit.REGISTRY.load(six.text_type(p), 'none')
        assert test.name == 'foobar'

    def test_folder_name(self, tmpdir):
        """backends.junit._load: uses the folder name if the result is
        'results.'
        """
        tmpdir.mkdir('foo')
        p = tmpdir.join('foo', 'results.xml')
        p.write(_XML)
        test = backends.junit.REGISTRY.load(six.text_type(p), 'none')

        assert test.name == 'foo'

    class TestReturned(object):
        """Test that the returned object is as expected."""

        testname = grouptools.join('foo', 'bar', 'a-test')

        @pytest.fixture
        def result(self, tmpdir):
            p = tmpdir.join('test.xml')
            p.write(_XML)
            return backends.junit._load(six.text_type(p))

        def test_testrunresult(self, result):
            """backends.junit._load: returns a TestrunResult instance."""
            assert isinstance(result, results.TestrunResult)

        def test_replace_sep(self, result):
            """backends.junit._load: replaces '.' with grouptools.SEPARATOR."""
            assert self.testname in result.tests

        def test_testresult_instance(self, result):
            """backends.junit._load: replaces result with TestResult instance.
            """
            assert isinstance(result.tests[self.testname], results.TestResult)

        def test_status_instance(self, result):
            """backends.junit._load: a status is found and loaded."""
            assert isinstance(result.tests[self.testname].result,
                              status.Status)

        def test_time(self, result):
            time = result.tests[self.testname].time
            assert isinstance(time, results.TimeAttribute)
            assert time.start == 1.0
            assert time.end == 4.5

        def test_command(self, result):
            """backends.junit._load: command is loaded correctly."""
            assert result.tests[self.testname].command == 'this/is/a/command'

        def test_out(self, result):
            """backends.junit._load: stdout is loaded correctly."""
            assert result.tests[self.testname].out == 'This is stdout'

        def test_err(self, result):
            """backends.junit._load: stderr is loaded correctly."""
            expected = textwrap.dedent("""\
                this is stderr

                pid: [1934]
                time start: 1.0
                time end: 4.5""")
            assert result.tests[self.testname].err.strip() == expected

        def test_totals(self, result):
            """backends.junit._load: Totals are calculated."""
            assert bool(result)

        def test_pid(self, result):
            """backends.junit._load: pid is loaded correctly."""
            assert result.tests[self.testname].pid == [1934]


class TestJUnitBackend(object):
    """Tests for the JUnitBackend class."""

    class TestFinalize(object):
        """Tests for the finalize method."""

        def test_skips_illformed_tests(self, tmpdir):
            """backends.junit.JUnitBackend: skips illformed tests"""
            result = results.TestResult()
            result.time.end = 1.2345
            result.result = 'pass'
            result.out = 'this is stdout'
            result.err = 'this is stderr'
            result.command = 'foo'

            test = backends.junit.JUnitBackend(six.text_type(tmpdir))
            test.initialize(shared.INITIAL_METADATA)
            with test.write_test(grouptools.join('a', 'group', 'test1')) as t:
                t(result)
            tmpdir.join('tests', '1.xml').write('bad data')

            test.finalize()


class TestJUnitWriter(object):
    """Tests for the JUnitWriter class."""

    def test_junit_replace(self, tmpdir):
        """backends.junit.JUnitBackend.write_test: grouptools.SEPARATOR is
        replaced with '.'.
        """
        result = results.TestResult()
        result.time.end = 1.2345
        result.result = 'pass'
        result.out = 'this is stdout'
        result.err = 'this is stderr'
        result.command = 'foo'

        test = backends.junit.JUnitBackend(six.text_type(tmpdir))
        test.initialize(shared.INITIAL_METADATA)
        with test.write_test(grouptools.join('a', 'group', 'test1')) as t:
            t(result)
        test.finalize()

        test_value = etree.parse(six.text_type(tmpdir.join('results.xml')))
        test_value = test_value.getroot()

        assert test_value.find('.//testcase').attrib['classname'] == \
            'piglit.a.group'

    class TestValid(object):
        @pytest.fixture
        def test_file(self, tmpdir):
            tmpdir.mkdir('foo')
            p = tmpdir.join('foo')

            result = results.TestResult()
            result.time.end = 1.2345
            result.result = 'pass'
            result.out = 'this is stdout'
            result.err = 'this is stderr'
            result.command = 'foo'
            result.pid = 1034

            test = backends.junit.JUnitBackend(six.text_type(p))
            test.initialize(shared.INITIAL_METADATA)
            with test.write_test(grouptools.join('a', 'group', 'test1')) as t:
                t(result)

            result.result = 'fail'
            with test.write_test(grouptools.join('a', 'test', 'test1')) as t:
                t(result)
            test.finalize()

            return six.text_type(p.join('results.xml'))

        def test_xml_well_formed(self, test_file):
            """backends.junit.JUnitBackend.write_test: produces well formed xml."""
            etree.parse(test_file)

        @pytest.mark.skipif(etree.__name__ != 'lxml.etree',
                            reason="This test requires lxml")
        def test_xml_valid(self, test_file):
            """backends.junit.JUnitBackend.write_test: produces valid JUnit xml."""
            # This XMLSchema class is unique to lxml
            schema = etree.XMLSchema(file=JUNIT_SCHEMA)  # pylint: disable=no-member
            with open(test_file, 'r') as f:
                assert schema.validate(etree.parse(f))


class TestJUnitSubtestWriter(object):
    """Tests for the JUnitWriter class."""

    def test_junit_replace(self, tmpdir):
        """backends.junit.JUnitBackend.write_test: grouptools.SEPARATOR is
        replaced with '.'.
        """
        result = results.TestResult()
        result.time.end = 1.2345
        result.out = 'this is stdout'
        result.err = 'this is stderr'
        result.command = 'foo'
        result.subtests['foo'] = 'pass'
        result.subtests['bar'] = 'fail'

        test = backends.junit.JUnitBackend(six.text_type(tmpdir),
                                           junit_subtests=True)
        test.initialize(shared.INITIAL_METADATA)
        with test.write_test(grouptools.join('a', 'group', 'test1')) as t:
            t(result)
        test.finalize()

        test_value = etree.parse(six.text_type(tmpdir.join('results.xml')))
        test_value = test_value.getroot()

        assert test_value.find('.//testsuite/testsuite').attrib['name'] == \
            'piglit.a.group.test1'

    def test_junit_replace_suffix(self, tmpdir):
        """backends.junit.JUnitBackend.write_test: grouptools.SEPARATOR is
        replaced with '.'.
        """
        result = results.TestResult()
        result.time.end = 1.2345
        result.out = 'this is stdout'
        result.err = 'this is stderr'
        result.command = 'foo'
        result.subtests['foo'] = 'pass'
        result.subtests['bar'] = 'fail'

        test = backends.junit.JUnitBackend(six.text_type(tmpdir),
                                           junit_subtests=True,
                                           junit_suffix='.foo')
        test.initialize(shared.INITIAL_METADATA)
        with test.write_test(grouptools.join('a', 'group', 'test1')) as t:
            t(result)
        test.finalize()

        test_value = etree.parse(six.text_type(tmpdir.join('results.xml')))
        test_value = test_value.getroot()

        suite = test_value.find('.//testsuite/testsuite')
        assert suite.attrib['name'] == 'piglit.a.group.test1'
        assert suite.find('.//testcase[@name="{}"]'.format('foo.foo')) is not None

    def test_subtest_skip(self, tmpdir):
        result = results.TestResult()
        result.time.end = 1.2345
        result.out = 'this is stdout'
        result.err = 'this is stderr'
        result.command = 'foo'
        result.subtests['foo'] = 'pass'
        result.subtests['bar'] = 'skip'

        test = backends.junit.JUnitBackend(six.text_type(tmpdir),
                                           junit_subtests=True)
        test.initialize(shared.INITIAL_METADATA)
        with test.write_test(grouptools.join('a', 'group', 'test1')) as t:
            t(result)
        test.finalize()

        test_value = etree.parse(six.text_type(tmpdir.join('results.xml')))
        test_value = test_value.getroot()

        suite = test_value.find('.//testsuite/testsuite')
        assert suite.attrib['name'] == 'piglit.a.group.test1'
        assert suite.find('.//testcase[@name="{}"]/skipped'.format('bar')) \
            is not None

    def test_result_skip(self, tmpdir):
        result = results.TestResult()
        result.time.end = 1.2345
        result.out = 'this is stdout'
        result.err = 'this is stderr'
        result.command = 'foo'
        result.result = 'skip'

        test = backends.junit.JUnitBackend(six.text_type(tmpdir),
                                           junit_subtests=True)
        test.initialize(shared.INITIAL_METADATA)
        with test.write_test(grouptools.join('a', 'group', 'test1')) as t:
            t(result)
        test.finalize()

        test_value = etree.parse(six.text_type(tmpdir.join('results.xml')))
        test_value = test_value.getroot()

        elem = test_value.find('.//testsuite/testcase[@name="test1"]/skipped')
        assert elem is not None

    def test_classname(self, tmpdir):
        result = results.TestResult()
        result.time.end = 1.2345
        result.out = 'this is stdout'
        result.err = 'this is stderr'
        result.command = 'foo'
        result.subtests['foo'] = 'pass'
        result.subtests['bar'] = 'skip'

        test = backends.junit.JUnitBackend(six.text_type(tmpdir),
                                           junit_subtests=True)
        test.initialize(shared.INITIAL_METADATA)
        with test.write_test(grouptools.join('a', 'group', 'test1')) as t:
            t(result)
        test.finalize()

        test_value = etree.parse(six.text_type(tmpdir.join('results.xml')))
        test_value = test_value.getroot()

        suite = test_value.find('.//testsuite/testsuite')
        assert suite.find('.//testcase[@classname="piglit.a.group.test1"]') \
            is not None

    class TestValid(object):
        @pytest.fixture
        def test_file(self, tmpdir):
            tmpdir.mkdir('foo')
            p = tmpdir.join('foo')

            result = results.TestResult()
            result.time.end = 1.2345
            result.out = 'this is stdout'
            result.err = 'this is stderr'
            result.command = 'foo'
            result.pid = 1034
            result.subtests['foo'] = 'pass'
            result.subtests['bar'] = 'fail'

            test = backends.junit.JUnitBackend(six.text_type(p),
                                               junit_subtests=True)
            test.initialize(shared.INITIAL_METADATA)
            with test.write_test(grouptools.join('a', 'group', 'test1')) as t:
                t(result)

            result.result = 'fail'
            with test.write_test(grouptools.join('a', 'test', 'test1')) as t:
                t(result)
            test.finalize()

            return six.text_type(p.join('results.xml'))

        def test_xml_well_formed(self, test_file):
            """backends.junit.JUnitBackend.write_test: produces well formed xml."""
            etree.parse(test_file)

        @pytest.mark.skipif(etree.__name__ != 'lxml.etree',
                            reason="This test requires lxml")
        def test_xml_valid(self, test_file):
            """backends.junit.JUnitBackend.write_test: produces valid JUnit xml."""
            # This XMLSchema class is unique to lxml
            schema = etree.XMLSchema(file=JUNIT_SCHEMA)  # pylint: disable=no-member
            with open(test_file, 'r') as f:
                assert schema.validate(etree.parse(f))



# coding=utf-8
# Copyright 2014-2016 Intel Corporation
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Piglit integrations for the official Khronos Vulkan CTS.

upstream: https://github.com/KhronosGroup/Vulkan-CTS

"""

from __future__ import (
    absolute_import, division, print_function, unicode_literals
)
import re

from framework.test import deqp

__all__ = ['profile']

# Path to the deqp-gles3 executable.
_DEQP_VK_BIN = deqp.get_option('PIGLIT_DEQP_VK_BIN',
                               ('deqp-vk', 'bin'),
                               required=True)

_EXTRA_ARGS = deqp.get_option('PIGLIT_DEQP_VK_EXTRA_ARGS',
                              ('deqp-vk', 'extra_args'),
                              default='').split()

_DEQP_ASSERT = re.compile(
    r'deqp-vk: external/vulkancts/.*: Assertion `.*\' failed.')


class DEQPVKTest(deqp.DEQPBaseTest):
    """Test representation for Khronos Vulkan CTS."""
    timeout = 60
    deqp_bin = _DEQP_VK_BIN
    @property
    def extra_args(self):
        return super(DEQPVKTest, self).extra_args + \
            [x for x in _EXTRA_ARGS if not x.startswith('--deqp-case')]

    def interpret_result(self):
        if 'Failed to compile shader at vkGlslToSpirV' in self.result.out:
            self.result.result = 'skip'
            self.result.out += \
                '\n\nMarked as skip because GLSLang failed to compile shaders'
        elif _DEQP_ASSERT.search(self.result.err):
            self.result.result = 'skip'
            self.result.out += \
                '\n\nMarked as skip because of a internal dEQP assertion'
        else:
            super(DEQPVKTest, self).interpret_result()


profile = deqp.make_profile(  # pylint: disable=invalid-name
    deqp.iter_deqp_test_cases(
        deqp.gen_caselist_txt(_DEQP_VK_BIN, 'dEQP-VK-cases.txt',
                              _EXTRA_ARGS)),
    DEQPVKTest)
